"""Project discovery utilities."""

from __future__ import annotations

import os
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional

from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.models import ProjectConfiguration


@dataclass
class ResolvedProject:
    """A resolved project directory with its label."""

    label: str
    project_dir: Path


class ProjectMarkerError(RuntimeError):
    """Raised when project discovery fails."""


def discover_project_directories(root: Path) -> List[Path]:
    """Discover project directories beneath the current root.

    :param root: Root directory to search from.
    :type root: Path
    :return: List of discovered project directories.
    :rtype: List[Path]
    :raises ProjectMarkerError: If a configured project path is missing.
    """
    project_dirs: List[Path] = []
    _collect_project_directories(root, project_dirs)
    project_dirs.extend(discover_kanbus_projects(root))
    project_dirs = _apply_ignore_paths(root, project_dirs)
    return _normalize_project_directories(project_dirs)


def discover_kanbus_projects(root: Path) -> List[Path]:
    """Discover project directories from Kanbus configuration only.

    :param root: Root directory to search from.
    :type root: Path
    :return: List of configured project directories.
    :rtype: List[Path]
    :raises ProjectMarkerError: If a referenced path is missing.
    """
    marker = _find_configuration_file(root)
    if marker is None:
        return []
    try:
        configuration = _load_configuration(marker)
    except RuntimeError as error:
        raise ProjectMarkerError(str(error)) from error
    project_dirs = _resolve_project_directories(marker.parent, configuration)
    return _normalize_project_directories(project_dirs)


def resolve_labeled_projects(root: Path) -> List[ResolvedProject]:
    """Resolve all labeled project directories from configuration.

    :param root: Repository root.
    :type root: Path
    :return: List of resolved projects with labels.
    :rtype: List[ResolvedProject]
    :raises ProjectMarkerError: If configuration or paths are invalid.
    """
    config_path = get_configuration_path(root)
    try:
        configuration = load_project_configuration(config_path)
    except RuntimeError as error:
        raise ProjectMarkerError(str(error)) from error
    return _resolve_labeled_project_directories(config_path.parent, configuration)


def _resolve_labeled_project_directories(
    base: Path, configuration: ProjectConfiguration
) -> List[ResolvedProject]:
    projects: List[ResolvedProject] = []
    primary = base / configuration.project_directory
    projects.append(
        ResolvedProject(label=configuration.project_key, project_dir=primary)
    )
    for label, vp in configuration.virtual_projects.items():
        candidate = Path(vp.path)
        if not candidate.is_absolute():
            candidate = base / candidate
        candidate = resolve_project_path(candidate)
        if not candidate.is_dir():
            raise ProjectMarkerError(f"virtual project path not found: {candidate}")
        projects.append(ResolvedProject(label=label, project_dir=candidate))
    return projects


def _resolve_project_directories(
    base: Path, configuration: ProjectConfiguration
) -> List[Path]:
    paths: List[Path] = []
    primary = base / configuration.project_directory
    paths.append(primary)
    for vp in configuration.virtual_projects.values():
        candidate = Path(vp.path)
        if not candidate.is_absolute():
            candidate = base / candidate
        candidate = resolve_project_path(candidate)
        if not candidate.is_dir():
            raise ProjectMarkerError(f"virtual project path not found: {candidate}")
        paths.append(candidate)
    return paths


def _collect_project_directories(root: Path, projects: List[Path]) -> None:
    """Collect project directories without deep recursion.

    Scan the root and one level of children for a "project" directory. This
    avoids pulling in deep fixture projects while still supporting repos that
    organize multiple projects under immediate subdirectories.
    """
    try:
        entries = list(root.iterdir())
    except OSError as error:
        raise ProjectMarkerError(str(error)) from error
    for entry in entries:
        if not entry.is_dir():
            continue
        name = entry.name
        if name == "project":
            projects.append(entry)
            continue
        nested_project = entry / "project"
        if nested_project.is_dir():
            projects.append(nested_project)
        # Do not recurse into subdirectories; explicit configuration controls
        # additional project discovery.


def _apply_ignore_paths(root: Path, project_dirs: List[Path]) -> List[Path]:
    """Filter out project directories matching ignore_paths from configuration."""
    marker = _find_configuration_file(root)
    if marker is None:
        return project_dirs
    try:
        configuration = _load_configuration(marker)
    except RuntimeError:
        return project_dirs
    if not configuration.ignore_paths:
        return project_dirs
    base = marker.parent
    ignored = set()
    for pattern in configuration.ignore_paths:
        ignore_path = base / pattern
        try:
            ignored.add(ignore_path.resolve())
        except OSError:
            pass
    return [path for path in project_dirs if path.resolve() not in ignored]


def _normalize_project_directories(paths: Iterable[Path]) -> List[Path]:
    normalized: List[Path] = []
    seen: set[Path] = set()
    for path in paths:
        candidate = resolve_project_path(path)
        if candidate in seen:
            continue
        seen.add(candidate)
        normalized.append(candidate)
    return sorted(normalized, key=lambda item: str(item))


def resolve_project_path(path: Path) -> Path:
    """Resolve a project path while tolerating filesystem errors.

    :param path: Path to resolve.
    :type path: Path
    :return: Resolved path or original path on failure.
    :rtype: Path
    """
    try:
        return _resolve_path(path)
    except OSError:
        return path


def _resolve_path(path: Path) -> Path:
    if os.getenv("KANBUS_TEST_CANONICALIZE_FAILURE"):
        raise OSError("forced canonicalize failure")
    return path.resolve()


def _find_configuration_file(root: Path) -> Optional[Path]:
    git_root = _find_git_root(root)
    current = root.resolve()
    while True:
        candidate = current / ".kanbus.yml"
        if candidate.is_file():
            return candidate
        if git_root is not None and current == git_root:
            break
        if current.parent == current:
            break
        current = current.parent
    return None


def _load_configuration(marker: Path) -> ProjectConfiguration:
    return load_project_configuration(marker)


def _find_git_root(root: Path) -> Optional[Path]:
    result = subprocess.run(
        ["git", "rev-parse", "--show-toplevel"],
        cwd=root,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        return None
    path = Path(result.stdout.strip())
    if path.is_dir():
        return path
    return None


def load_project_directory(root: Path) -> Path:
    """Load the primary project directory for write operations.

    When a configuration file is found, derives the project directory directly
    from ``project_directory`` in the config. Virtual project directories are
    for reading and lookup only and are not returned here.

    :param root: Repository root path.
    :type root: Path
    :return: Path to the primary project directory.
    :rtype: Path
    :raises ProjectMarkerError: If no project is found or the configuration
        is invalid.
    """
    marker = _find_configuration_file(root)
    if marker is not None:
        try:
            configuration = _load_configuration(marker)
        except RuntimeError as error:
            raise ProjectMarkerError(str(error)) from error
        primary = marker.parent / configuration.project_directory
        filtered = _apply_ignore_paths(root, [primary])
        if not filtered:
            raise ProjectMarkerError("project not initialized")
        return _normalize_project_directories(filtered)[0]

    # No config file — fall back to filesystem scanning.
    project_dirs: List[Path] = []
    _collect_project_directories(root, project_dirs)
    project_dirs = _apply_ignore_paths(root, project_dirs)
    project_dirs = _normalize_project_directories(project_dirs)
    if not project_dirs:
        raise ProjectMarkerError("project not initialized")
    if len(project_dirs) > 1:
        discovered = ", ".join(str(path) for path in project_dirs)
        raise ProjectMarkerError(
            f"multiple projects found: {discovered}. "
            "Run this command from a directory with a single project/"
        )
    return project_dirs[0]


def get_configuration_path(root: Path) -> Path:
    """Return the configuration file path.

    :param root: Repository root path.
    :type root: Path
    :return: Path to .kanbus.yml.
    :rtype: Path
    :raises ProjectMarkerError: If the configuration file is missing.
    :raises ConfigurationError: If configuration path lookup fails.
    """
    if os.getenv("KANBUS_TEST_CONFIGURATION_PATH_FAILURE"):
        raise ConfigurationError("configuration path lookup failed")
    marker = _find_configuration_file(root)
    if marker is None:
        raise ProjectMarkerError("project not initialized")
    return marker


def find_project_local_directory(project_dir: Path) -> Optional[Path]:
    """Find a sibling project-local directory for a project.

    :param project_dir: Shared project directory.
    :type project_dir: Path
    :return: Project-local directory if present.
    :rtype: Optional[Path]
    """
    local_dir = project_dir.parent / "project-local"
    if local_dir.is_dir():
        return local_dir
    return None


def ensure_project_local_directory(project_dir: Path) -> Path:
    """Ensure the project-local directory exists and is gitignored.

    :param project_dir: Shared project directory.
    :type project_dir: Path
    :return: Path to the project-local directory.
    :rtype: Path
    """
    local_dir = project_dir.parent / "project-local"
    issues_dir = local_dir / "issues"
    events_dir = local_dir / "events"
    issues_dir.mkdir(parents=True, exist_ok=True)
    events_dir.mkdir(parents=True, exist_ok=True)
    _ensure_gitignore_entry(project_dir.parent, "project-local/")
    return local_dir


def _ensure_gitignore_entry(root: Path, entry: str) -> None:
    gitignore_path = root / ".gitignore"
    if gitignore_path.exists():
        contents = gitignore_path.read_text(encoding="utf-8")
        lines = [line.strip() for line in contents.splitlines()]
    else:
        contents = ""
        lines = []
    if entry in lines:
        return
    suffix = "" if contents.endswith("\n") or contents == "" else "\n"
    gitignore_path.write_text(f"{contents}{suffix}{entry}\n", encoding="utf-8")



"""Workflow validation and transition side effects."""

from __future__ import annotations

from datetime import datetime
from typing import Dict, List

from kanbus.models import IssueData, ProjectConfiguration


class InvalidTransitionError(RuntimeError):
    """Raised when a workflow status transition is invalid."""


def get_workflow_for_issue_type(
    configuration: ProjectConfiguration,
    issue_type: str,
) -> Dict[str, List[str]]:
    """Return the workflow definition for a specific issue type.

    :param configuration: Project configuration with workflow definitions.
    :type configuration: ProjectConfiguration
    :param issue_type: Issue type to lookup.
    :type issue_type: str
    :return: Workflow definition for the issue type.
    :rtype: Dict[str, List[str]]
    :raises ValueError: If the default workflow is missing.
    """
    workflows = configuration.workflows
    if issue_type in workflows:
        return workflows[issue_type]
    if "default" not in workflows:
        raise ValueError("default workflow not defined")
    return workflows["default"]


def validate_status_transition(
    configuration: ProjectConfiguration,
    issue_type: str,
    current_status: str,
    new_status: str,
) -> None:
    """Validate that a status transition is permitted by the workflow.

    Looks up the workflow for the given issue type in the project
    configuration (falling back to the default workflow if no
    type-specific workflow exists), then verifies that the new status
    appears in the list of allowed transitions from the current status.

    :param configuration: Project configuration containing workflow definitions.
    :type configuration: ProjectConfiguration
    :param issue_type: Issue type being transitioned.
    :type issue_type: str
    :param current_status: Issue's current status.
    :type current_status: str
    :param new_status: Desired new status.
    :type new_status: str
    :raises InvalidTransitionError: If the transition is not permitted.
    """
    workflow = get_workflow_for_issue_type(configuration, issue_type)
    allowed_transitions = workflow.get(current_status, [])
    if new_status not in allowed_transitions:
        raise InvalidTransitionError(
            f"invalid transition from '{current_status}' "
            f"to '{new_status}' for type '{issue_type}'"
        )


def _collect_workflow_statuses(workflow: Dict[str, List[str]]) -> set[str]:
    valid_statuses = set(workflow.keys())
    for allowed in workflow.values():
        valid_statuses.update(allowed)
    return valid_statuses


def validate_status_value(
    configuration: ProjectConfiguration, issue_type: str, status: str
) -> None:
    valid_statuses = {s.key for s in configuration.statuses}
    if status not in valid_statuses:
        raise InvalidTransitionError("unknown status")

    workflow = get_workflow_for_issue_type(configuration, issue_type)
    if status not in _collect_workflow_statuses(workflow):
        raise InvalidTransitionError("invalid transition")


def apply_transition_side_effects(
    issue: IssueData,
    new_status: str,
    current_utc_time: datetime,
) -> IssueData:
    """Apply workflow side effects based on a status transition.

    :param issue: Issue being updated.
    :type issue: IssueData
    :param new_status: New status being applied.
    :type new_status: str
    :param current_utc_time: Current UTC timestamp.
    :type current_utc_time: datetime
    :return: Updated issue data with side effects applied.
    :rtype: IssueData
    """
    closed_at = issue.closed_at
    if new_status == "closed":
        closed_at = current_utc_time
    elif issue.status == "closed" and new_status != "closed":
        closed_at = None
    return issue.model_copy(update={"closed_at": closed_at})



"""Issue deletion workflow."""

from __future__ import annotations

from pathlib import Path

from kanbus.issue_files import write_issue_to_file
from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
from kanbus.event_history import (
    create_event,
    events_dir_for_issue_path,
    issue_deleted_payload,
    now_timestamp,
    write_events_batch,
)
from kanbus.users import get_current_user


class IssueDeleteError(RuntimeError):
    """Raised when issue deletion fails."""


def delete_issue(root: Path, identifier: str) -> None:
    """Delete an issue file from disk.

    :param root: Repository root path.
    :type root: Path
    :param identifier: Issue identifier.
    :type identifier: str
    :raises IssueDeleteError: If deletion fails.
    """
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueDeleteError(str(error)) from error

    lookup.issue_path.unlink()
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=lookup.issue.identifier,
        event_type="issue_deleted",
        actor_id=actor_id,
        payload=issue_deleted_payload(lookup.issue),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(lookup.project_dir, lookup.issue_path)
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(lookup.issue, lookup.issue_path)
        raise IssueDeleteError(str(error)) from error



"""Environment diagnostics for Kanbus."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.file_io import InitializationError, ensure_git_repository
from kanbus.project import (
    ProjectMarkerError,
    get_configuration_path,
    load_project_directory,
)


class DoctorError(RuntimeError):
    """Raised when doctor checks fail."""


@dataclass(frozen=True)
class DoctorResult:
    """Result of running doctor checks."""

    project_dir: Path


def run_doctor(root: Path) -> DoctorResult:
    """Run diagnostic checks for Kanbus.

    :param root: Repository root path.
    :type root: Path
    :return: Doctor result with project directory.
    :rtype: DoctorResult
    :raises DoctorError: If any check fails.
    """
    try:
        ensure_git_repository(root)
    except InitializationError as error:
        raise DoctorError(str(error)) from error

    try:
        project_dir = load_project_directory(root)
    except ProjectMarkerError as error:
        raise DoctorError(str(error)) from error

    try:
        load_project_configuration(get_configuration_path(project_dir))
    except ConfigurationError as error:
        raise DoctorError(str(error)) from error

    return DoctorResult(project_dir=project_dir)



"""Kanbus data models."""

from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field


class CategoryDefinition(BaseModel):
    """Category definition for grouping statuses."""

    name: str = Field(min_length=1)
    color: Optional[str] = None


class DependencyLink(BaseModel):
    """Dependency link between issues.

    :param target: Target issue identifier.
    :type target: str
    :param dependency_type: Dependency type.
    :type dependency_type: str
    """

    target: str = Field(min_length=1)
    dependency_type: str = Field(alias="type", min_length=1)


class IssueComment(BaseModel):
    """Comment on an issue.

    :param author: Comment author.
    :type author: str
    :param text: Comment text.
    :type text: str
    :param created_at: Timestamp when the comment was created.
    :type created_at: datetime
    """

    id: Optional[str] = None
    author: str = Field(min_length=1)
    text: str = Field(min_length=1)
    created_at: datetime


class IssueData(BaseModel):
    """Issue data representation.

    :param identifier: Issue ID.
    :type identifier: str
    :param title: Short summary.
    :type title: str
    :param description: Markdown body.
    :type description: str
    :param issue_type: Issue type.
    :type issue_type: str
    :param status: Issue status.
    :type status: str
    :param priority: Priority level.
    :type priority: int
    :param assignee: Assignee identifier.
    :type assignee: Optional[str]
    :param creator: Creator identifier.
    :type creator: Optional[str]
    :param parent: Parent issue identifier.
    :type parent: Optional[str]
    :param labels: Labels for the issue.
    :type labels: List[str]
    :param dependencies: Dependency links.
    :type dependencies: List[DependencyLink]
    :param comments: Issue comments.
    :type comments: List[IssueComment]
    :param created_at: Creation timestamp.
    :type created_at: datetime
    :param updated_at: Update timestamp.
    :type updated_at: datetime
    :param closed_at: Close timestamp.
    :type closed_at: Optional[datetime]
    :param custom: Custom fields.
    :type custom: Dict[str, object]
    """

    identifier: str = Field(alias="id", min_length=1)
    title: str = Field(min_length=1)
    description: str = ""
    issue_type: str = Field(alias="type", min_length=1)
    status: str = Field(min_length=1)
    priority: int
    assignee: Optional[str] = None
    creator: Optional[str] = None
    parent: Optional[str] = None
    labels: List[str] = Field(default_factory=list)
    dependencies: List[DependencyLink] = Field(default_factory=list)
    comments: List[IssueComment] = Field(default_factory=list)
    created_at: datetime
    updated_at: datetime
    closed_at: Optional[datetime] = None
    custom: Dict[str, object] = Field(default_factory=dict)


class StatusDefinition(BaseModel):
    """Status definition with display metadata."""

    key: str = Field(min_length=1)
    name: str = Field(min_length=1)
    category: str = Field(min_length=1)
    color: Optional[str] = None
    collapsed: bool = False


class PriorityDefinition(BaseModel):
    """Priority definition containing label and optional color."""

    name: str = Field(min_length=1)
    color: Optional[str] = None


class JiraConfiguration(BaseModel):
    """Jira synchronization configuration."""

    model_config = ConfigDict(extra="forbid")

    url: str
    project_key: str
    sync_direction: str = "pull"
    type_mappings: Dict[str, str] = Field(default_factory=dict)
    field_mappings: Dict[str, str] = Field(default_factory=dict)


class SnykConfiguration(BaseModel):
    """Snyk vulnerability synchronization configuration."""

    model_config = ConfigDict(extra="forbid")

    org_id: str
    min_severity: str = "low"
    parent_epic: Optional[str] = None
    repo: Optional[str] = None


class VirtualProjectConfig(BaseModel):
    """Configuration for a single virtual project."""

    model_config = ConfigDict(extra="forbid")

    path: str


class ProjectConfiguration(BaseModel):
    """Project configuration loaded from .kanbus.yml.

    :param project_directory: Relative path to the primary project directory.
    :type project_directory: str
    :param virtual_projects: Map of label to virtual project configuration.
    :type virtual_projects: Dict[str, VirtualProjectConfig]
    :param new_issue_project: Project label for new issue creation routing.
    :type new_issue_project: Optional[str]
    :param ignore_paths: Paths to exclude from project discovery.
    :type ignore_paths: List[str]
    :param project_key: Issue ID project key (prefix).
    :type project_key: str
    :param project_management_template: Optional template path for CONTRIBUTING_AGENT.md.
    :type project_management_template: Optional[str]
    :param hierarchy: Hierarchy ordering.
    :type hierarchy: List[str]
    :param types: Non-hierarchical types.
    :type types: List[str]
    :param workflows: Workflow definitions.
    :type workflows: Dict[str, Dict[str, List[str]]]
    :param initial_status: Initial status for new issues.
    :type initial_status: str
    :param priorities: Priority map.
    :type priorities: Dict[int, PriorityDefinition]
    :param default_priority: Default priority.
    :type default_priority: int
    :param assignee: Default assignee identifier.
    :type assignee: Optional[str]
    :param time_zone: Preferred display time zone.
    :type time_zone: Optional[str]
    :param type_colors: Optional map of issue type to color name.
    :type type_colors: Dict[str, str]
    :param beads_compatibility: Default Beads compatibility mode.
    :type beads_compatibility: bool
    :param jira: Optional Jira synchronization configuration.
    :type jira: Optional[JiraConfiguration]
    :param snyk: Optional Snyk vulnerability synchronization configuration.
    :type snyk: Optional[SnykConfiguration]
    """

    model_config = ConfigDict(extra="forbid")

    project_directory: str
    virtual_projects: Dict[str, VirtualProjectConfig] = Field(default_factory=dict)
    new_issue_project: Optional[str] = None
    ignore_paths: List[str] = Field(default_factory=list)
    console_port: Optional[int] = None
    project_key: str = Field(min_length=1)
    project_management_template: Optional[str] = None
    hierarchy: List[str]
    types: List[str]
    workflows: Dict[str, Dict[str, List[str]]]
    transition_labels: Dict[str, Dict[str, Dict[str, str]]] = Field(
        default_factory=dict
    )
    initial_status: str = Field(min_length=1)
    priorities: Dict[int, PriorityDefinition]
    default_priority: int
    assignee: Optional[str] = Field(default=None, min_length=1)
    time_zone: Optional[str] = Field(default=None, min_length=1)
    statuses: List[StatusDefinition] = Field(default_factory=list)
    categories: List[CategoryDefinition] = Field(default_factory=list)
    type_colors: Dict[str, str] = Field(default_factory=dict)
    beads_compatibility: bool = False
    jira: Optional[JiraConfiguration] = None
    snyk: Optional[SnykConfiguration] = None



"""Policy evaluation engine."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from kanbus.policy_context import PolicyContext, PolicyViolationError
from kanbus.policy_steps import StepOutcome, StepRegistry

if TYPE_CHECKING:
    from gherkin.ast import GherkinDocument, Step


class ScenarioResult(str, Enum):
    """Result of evaluating a scenario."""

    PASSED = "passed"
    SKIPPED = "skipped"


_REGISTRY: StepRegistry | None = None


def _get_step_registry() -> StepRegistry:
    global _REGISTRY
    if _REGISTRY is None:
        _REGISTRY = StepRegistry()
    return _REGISTRY


@dataclass
class PolicyEvaluationOptions:
    """Options for policy evaluation."""

    collect_all_violations: bool = False


def evaluate_policies(
    context: PolicyContext, documents: list[tuple[str, GherkinDocument]]
) -> None:
    """Evaluate all policies against the given context.

    :param context: Policy evaluation context.
    :type context: PolicyContext
    :param documents: List of tuples containing filename and parsed Gherkin document.
    :type documents: list[tuple[str, GherkinDocument]]
    :raises PolicyViolationError: If any policy rule fails.
    """
    violations = evaluate_policies_with_options(
        context, documents, PolicyEvaluationOptions()
    )
    if violations:
        raise violations[0]


def evaluate_policies_with_options(
    context: PolicyContext,
    documents: list[tuple[str, GherkinDocument]],
    options: PolicyEvaluationOptions,
) -> list[PolicyViolationError]:
    """Evaluate all policies and optionally collect all violations.

    :param context: Policy evaluation context.
    :type context: PolicyContext
    :param documents: List of tuples containing filename and parsed Gherkin document.
    :type documents: list[tuple[str, GherkinDocument]]
    :param options: Evaluation options.
    :type options: PolicyEvaluationOptions
    :return: List of policy violations.
    :rtype: list[PolicyViolationError]
    """
    registry = _get_step_registry()
    violations = []

    for filename, document in documents:
        if not document.feature:
            continue

        for child in document.feature.children:
            if not hasattr(child, "scenario") or not child.scenario:
                continue

            scenario = child.scenario
            try:
                evaluate_scenario(
                    context, registry, filename, scenario.name, scenario.steps
                )
            except PolicyViolationError as error:
                violations.append(error)
                if not options.collect_all_violations:
                    return violations

    return violations


def evaluate_scenario(
    context: PolicyContext,
    registry: StepRegistry,
    policy_file: str,
    scenario_name: str,
    steps: list[Step],
) -> ScenarioResult:
    """Evaluate a single scenario.

    :param context: Policy evaluation context.
    :type context: PolicyContext
    :param registry: Step registry.
    :type registry: StepRegistry
    :param policy_file: Name of the policy file.
    :type policy_file: str
    :param scenario_name: Name of the scenario.
    :type scenario_name: str
    :param steps: List of steps in the scenario.
    :type steps: list[Step]
    :return: Scenario result.
    :rtype: ScenarioResult
    :raises PolicyViolationError: If any Then step fails.
    """
    scenario_skipped = False

    for step in steps:
        step_text = step.text.strip()
        step_keyword = step.keyword.strip()

        found = registry.find_step(step_text)
        if not found:
            raise PolicyViolationError(
                policy_file=policy_file,
                scenario=scenario_name,
                failed_step=f"{step_keyword} {step_text}",
                message=f"no matching step definition for: {step_text}",
                issue_id=context.proposed_issue.identifier,
            )

        step_def, match = found
        outcome, error_message = step_def.execute(context, match)

        if outcome == StepOutcome.PASS:
            continue

        if outcome == StepOutcome.FAIL:
            raise PolicyViolationError(
                policy_file=policy_file,
                scenario=scenario_name,
                failed_step=f"{step_keyword} {step_text}",
                message=error_message or "step failed without explanation",
                issue_id=context.proposed_issue.identifier,
            )

        if outcome == StepOutcome.SKIP:
            if step_keyword in ("Given", "When"):
                scenario_skipped = True
                break
            continue

    if scenario_skipped:
        return ScenarioResult.SKIPPED
    return ScenarioResult.PASSED



"""Default CONTRIBUTING_AGENT template content."""

from __future__ import annotations

DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME = "CONTRIBUTING_AGENT.template.md"

DEFAULT_PROJECT_MANAGEMENT_TEMPLATE = """# THE WAY

This repository is governed by The Way.
The Way is not a suggestion. It is the prescribed order of creation.
All work must pass through its forms.
This is The Way.

Deviation from The Way is extremely offensive to the religious sensibilities of its practitioners, including the human user of this coding agent. Sins against The Way must be avoided at all costs.

## The Sacrament of Work

The Sacrament of Work is the sacrament of vibe coding in this repository.

Kanbus is the instrument of record. Nothing exists until it is named there.

Work begins in declaration, not in code.
Code without declaration is disorder.

Every change must be:
- Named.
- Described.
- Specified.
- Proven.
- Recorded.

If it is not recorded, it did not happen.

Use Markdown in issue descriptions and comments whenever it helps make the intent clearer.
When relationships, flows, or state transitions are easier to grasp visually, add a Mermaid diagram.

Editing project/ directly is hacking the data and is a sin against The Way. Do not read or write anything inside project/. Do not inspect issue JSON with tools like cat or jq. All work must pass through Kanbus.

## The Order of Being

All work is structured.

Project key prefix: {{ project_key }}.

Hierarchy: {{ hierarchy_order }}.

Non-hierarchical types: {{ non_hierarchical_types | join(", ") if non_hierarchical_types else "none" }}.

Only hierarchy types may be parents.

Permitted relationships are fixed and not to be altered.

Allowed parent-child relationships:
{% for rule in parent_child_rules %}
- {{ rule }}
{% endfor %}

Structure is not bureaucracy. Structure is memory.

## The Cognitive Framework

There is one discipline.

Outside-in Behavior-Driven Design.

The specification is the product.
Production code exists only to make a failing specification pass.

This is the first principle.

Non-negotiable laws:
- Begin with intent, not internals.
- Describe behavior in English.
- Translate behavior into Gherkin.
- Run it and watch it fail.
- Write only the code required to make it pass.
- All behavior must be specified.
- No specification may be red.
- Specifications describe observable behavior only.
- Specifications must not describe internal structure.

If behavior cannot be observed, it is not behavior.

## Roles in the Order

Epics define purpose and completion.

Stories define behavior. They contain Gherkin. They define what must happen.

Tasks and sub-tasks define implementation. They may not invent behavior beyond the specification.

Bugs restore violated behavior.

Chores maintain the ground on which behavior stands.

## The Rite of Gherkin

Every story must contain a Gherkin form.

Minimum structure:
{% for line in gherkin_example %}
{{ line }}
{% endfor %}

This is required.

Without this form, there is no alignment between intent and implementation.

## The Outside-In Ritual

When asked to add or change behavior, follow this sequence. It is not optional.
1. Clarify intent in English.
Capture role, capability, benefit.
Use: As a <role>, I want <capability>, so that <benefit>.
Confirm what is not included.
2. Create the epic and stories in Kanbus.
Record intent and Definition of Done.
3. Write executable specifications before any production code.
4. Run the specifications and confirm they fail.
5. Write the smallest code necessary to pass.
6. Refactor only while all specifications remain green.
7. Record progress. Close only when complete.

Skipping steps is corruption of the process.

## Coverage

100% specification coverage is mandatory.

Every behavior must be specified.
Every specification must pass.

Green is peace. Red is unfinished.

## Status and Priority

Statuses and workflows are fixed. They exist to maintain order.

Initial status: {{ initial_status }}.
Status changes must follow the workflow transitions below.
Workflow selection: use a workflow named after the issue type when present; otherwise use the default workflow.

{% for workflow in workflows %}
{{ workflow.name }} workflow:
{% if workflow.statuses %}
{% for status in workflow.statuses %}
- {{ status.name }} -> {{ status.transitions | join(", ") if status.transitions else "none" }}
{% endfor %}
{% else %}
- No statuses defined.
{% endif %}

{% endfor %}
Priorities are:

{% for priority in priorities %}
- {{ priority.value }} -- {{ priority.name }}
{% endfor %}
Default is {{ default_priority_value }} ({{ default_priority_name }}).

Severity is not emotion. It is signal.

## Command examples

{% for command in command_examples %}
{{ command }}
{% endfor %}

## Semantic Release Alignment

Issue types map directly to release categories.

{% for mapping in semantic_release_mapping %}
- {{ mapping.type }} -> {{ mapping.category }}
{% endfor %}

Release notes are not commentary. They are a ledger of truth.

## Example: Hello World

Even the smallest program must pass through The Way.

No code precedes intent.
No intent precedes recording.
No implementation precedes failure.

The smallest program is still subject to discipline.

User request: "Please create a Hello World program."

1. Interview the stakeholder before any code
Ask why they want it and capture the intent in plain English.
Example prompts:
- Who is the audience for Hello World?
- What environment or language should it run in?
- What output is required and where should it appear?
- What is out of scope?

2. Convert intent into a user story (before any code)
Example:
As a new user, I want a Hello World program, so that I can verify the toolchain works.

3. Create an epic for the milestone and record the story
Command:
kanbus create "Hello World program" --type epic

Example output (capture the ID):
ID: kanbus-1a2b3c

Record the intent on the epic:
kanbus comment kanbus-1a2b3c "As a new user, I want a Hello World program, so that I can verify the toolchain works."

4. Create a story for the behavior and include Gherkin (before any code)
Command:
kanbus create "Prints Hello World to stdout" --type story --parent kanbus-1a2b3c

Example output (capture the story ID):
ID: kanbus-4d5e6f

Attach the Gherkin acceptance criteria:
kanbus comment kanbus-4d5e6f "Feature: Hello World
  Scenario: Run the program
    Given a configured environment
    When I run the program
    Then it prints \"Hello, world\" to stdout"

5. Run the Gherkin and confirm it fails (before any production code)
Run the behavior tests in the repo and confirm the new scenario fails for the right reason.

6. Implement the minimum code to pass, then refactor
Write the smallest change that makes the Gherkin scenario pass.
Refactor only while all specs remain green.
"""



"""Issue display formatting helpers."""

from __future__ import annotations

import os
import sys
from typing import Dict, Optional

import click

from kanbus.ids import format_issue_key
from kanbus.models import IssueData, ProjectConfiguration

STATUS_GLYPHS = {
    "backlog": "○",
    "open": "◌",
    "in_progress": "◐",
    "blocked": "◑",
    "closed": "●",
    "deferred": "◍",
}

DEFAULT_STATUS_COLORS: Dict[str, str] = {
    "backlog": "grey",
    "open": "cyan",
    "in_progress": "blue",
    "blocked": "red",
    "closed": "green",
    "deferred": "yellow",
}

DEFAULT_PRIORITY_COLORS: Dict[int, str] = {
    0: "red",
    1: "bright_red",
    2: "yellow",
    3: "blue",
    4: "white",
}

DEFAULT_TYPE_COLORS: Dict[str, str] = {
    "initiative": "bright_blue",
    "epic": "magenta",
    "task": "cyan",
    "sub-task": "bright_cyan",
    "bug": "red",
    "story": "yellow",
    "chore": "green",
    "event": "bright_blue",
}

KNOWN_COLORS = {
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
}


def _should_use_color() -> bool:
    context = click.get_current_context(silent=True)
    if context is not None and context.color is not None:
        return context.color
    return sys.stdout.isatty() and os.getenv("NO_COLOR") is None


def _dim(text: str, use_color: bool) -> str:
    if not use_color:
        return text
    return click.style(text, fg="bright_black")


def _normalize_color(color: Optional[str]) -> Optional[str]:
    return color if color in KNOWN_COLORS else None


def _paint(value: str, color: Optional[str], use_color: bool) -> str:
    if not use_color:
        return value
    normalized = _normalize_color(color)
    if normalized is None:
        return value
    return click.style(value, fg=normalized)


def format_issue_for_display(
    issue: IssueData,
    configuration: Optional[ProjectConfiguration] = None,
    use_color: Optional[bool] = None,
    project_context: bool = False,
) -> str:
    """Format an issue for human-readable display.

    :param issue: Issue data to display.
    :type issue: IssueData
    :param use_color: Whether to apply ANSI colors (defaults to TTY detection).
    :type use_color: Optional[bool]
    :param project_context: Whether the identifier should omit the project key.
    :type project_context: bool
    :return: Human-readable issue display.
    :rtype: str
    """
    color_output = _should_use_color() if use_color is None else use_color

    # Build status_colors from statuses list
    status_colors = DEFAULT_STATUS_COLORS.copy()
    if configuration:
        category_colors = {
            category.name: category.color for category in configuration.categories
        }
        for status_def in configuration.statuses:
            if status_def.color:
                status_colors[status_def.key] = status_def.color
            elif status_def.category and status_def.category in category_colors:
                color = category_colors[status_def.category]
                if color:
                    status_colors[status_def.key] = color
    priority_colors: Dict[int, str] = DEFAULT_PRIORITY_COLORS
    if configuration:
        priority_colors = priority_colors.copy()
        for value, definition in configuration.priorities.items():
            if definition.color:
                priority_colors[value] = definition.color
    type_colors = (
        {**DEFAULT_TYPE_COLORS, **configuration.type_colors}
        if configuration
        else DEFAULT_TYPE_COLORS
    )

    labels_text = ", ".join(issue.labels) if issue.labels else "-"

    formatted_identifier = format_issue_key(issue.identifier, project_context)

    rows = [
        ("ID:", formatted_identifier, None, False),
        ("Title:", issue.title, None, False),
        ("Type:", issue.issue_type, type_colors.get(issue.issue_type), False),
        ("Status:", issue.status, status_colors.get(issue.status), False),
        ("Priority:", str(issue.priority), priority_colors.get(issue.priority), False),
        ("Assignee:", issue.assignee or "-", None, issue.assignee is None),
        ("Parent:", issue.parent or "-", None, issue.parent is None),
        ("Labels:", labels_text, None, not bool(issue.labels)),
    ]

    lines = []
    for label, value, color, muted in rows:
        painted_value = _paint(
            value, color if not muted else "bright_black", color_output
        )
        lines.append(f"{_dim(label, color_output)} {painted_value}")

    if issue.description:
        lines.append(f"{_dim('Description:', color_output)}")
        lines.append(_paint(issue.description, None, color_output))

    if issue.dependencies:
        lines.append(f"{_dim('Dependencies:', color_output)}")
        for dependency in issue.dependencies:
            lines.append(f"  {dependency.dependency_type}: {dependency.target}")

    if issue.comments:
        lines.append(f"{_dim('Comments:', color_output)}")
        for comment in issue.comments:
            author = comment.author or "unknown"
            prefix = (comment.id or "")[:6]
            if prefix:
                lines.append(
                    f"  [{prefix}] {_dim(f'{author}:', color_output)} {comment.text}"
                )
            else:
                lines.append(f"  {_dim(f'{author}:', color_output)} {comment.text}")

    return "\n".join(lines)



"""Jira synchronization support.

Pulls issues from a remote Jira project into the local Kanbus project.
Secrets are read from environment variables JIRA_API_TOKEN and JIRA_USER_EMAIL.
"""

from __future__ import annotations

import os
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import requests

from kanbus.ids import IssueIdentifierRequest, generate_issue_identifier
from kanbus.issue_files import (
    list_issue_identifiers,
    read_issue_from_file,
    write_issue_to_file,
)
from kanbus.models import IssueComment, IssueData, JiraConfiguration


class JiraSyncError(RuntimeError):
    """Raised when a Jira sync operation fails."""


@dataclass
class JiraPullResult:
    """Result of a Jira pull operation."""

    pulled: int = 0
    updated: int = 0


def pull_from_jira(
    root: Path,
    jira_config: JiraConfiguration,
    project_key: str,
    dry_run: bool = False,
) -> JiraPullResult:
    """Pull issues from a Jira project into the local Kanbus project.

    :param root: Repository root path.
    :param jira_config: Jira configuration from .kanbus.yml.
    :param project_key: Kanbus project key (issue ID prefix).
    :param dry_run: If True, print what would be done without writing files.
    :raises JiraSyncError: If authentication or API calls fail.
    """
    api_token = os.environ.get("JIRA_API_TOKEN")
    user_email = os.environ.get("JIRA_USER_EMAIL")

    if not api_token:
        raise JiraSyncError("JIRA_API_TOKEN environment variable is not set")
    if not user_email:
        raise JiraSyncError("JIRA_USER_EMAIL environment variable is not set")

    from kanbus.project import load_project_directory

    project_dir = load_project_directory(root)
    issues_dir = project_dir / "issues"

    if not issues_dir.exists():
        raise JiraSyncError("issues directory does not exist")

    jira_issues = _fetch_all_jira_issues(jira_config, user_email, api_token)

    existing_ids = list_issue_identifiers(issues_dir)
    jira_key_index = _build_jira_key_index(existing_ids, issues_dir)

    # Pre-assign Kanbus IDs to new issues so parent links can be resolved
    jira_key_to_kanbus_id: Dict[str, str] = dict(jira_key_index)
    new_issue_ids: Dict[str, str] = {}
    all_existing: Set[str] = set(existing_ids)

    for jira_issue in jira_issues:
        jira_key = _jira_issue_key(jira_issue)
        if jira_key not in jira_key_to_kanbus_id:
            request = IssueIdentifierRequest(
                title=_jira_issue_summary(jira_issue),
                existing_ids=frozenset(all_existing),
                prefix=project_key,
            )
            result = generate_issue_identifier(request)
            all_existing.add(result.identifier)
            new_issue_ids[jira_key] = result.identifier
            jira_key_to_kanbus_id[jira_key] = result.identifier

    result_counts = JiraPullResult()

    for jira_issue in jira_issues:
        jira_key = _jira_issue_key(jira_issue)
        kanbus_issue = _map_jira_to_kanbus(
            jira_issue, jira_config, jira_key_to_kanbus_id
        )

        existing_kanbus_id = jira_key_index.get(jira_key)
        if existing_kanbus_id:
            kanbus_id = existing_kanbus_id
            action = "updated"
        else:
            kanbus_id = new_issue_ids[jira_key]
            action = "pulled "

        kanbus_issue = kanbus_issue.model_copy(update={"identifier": kanbus_id})

        # For updates, preserve fields not managed by Jira
        issue_path = issue_path_for_identifier(issues_dir, kanbus_id)
        if action == "updated" and issue_path.exists():
            try:
                existing = read_issue_from_file(issue_path)
                kanbus_issue = kanbus_issue.model_copy(
                    update={"created_at": existing.created_at}
                )
            except Exception:
                pass

        short_key = (
            kanbus_id[: kanbus_id.find("-") + 7] if "-" in kanbus_id else kanbus_id[:6]
        )
        print(f'{action}  {jira_key:<12}  {short_key:<14}  "{kanbus_issue.title}"')

        if not dry_run:
            write_issue_to_file(kanbus_issue, issue_path)

        if action == "updated":
            result_counts.updated += 1
        else:
            result_counts.pulled += 1

    return result_counts


def _fetch_all_jira_issues(
    jira_config: JiraConfiguration,
    user_email: str,
    api_token: str,
) -> List[Dict[str, Any]]:
    """Fetch all issues from Jira with pagination."""
    base_url = jira_config.url.rstrip("/")
    project_key = jira_config.project_key
    fields = (
        "summary,description,issuetype,status,priority,assignee,"
        "reporter,parent,labels,comment,created,updated,resolutiondate"
    )

    auth = (user_email, api_token)
    headers = {"Accept": "application/json"}
    all_issues: List[Dict[str, Any]] = []
    start_at = 0
    max_results = 100

    while True:
        url = (
            f"{base_url}/rest/api/3/search/jql"
            f"?jql=project={project_key}+ORDER+BY+created+ASC"
            f"&fields={fields}&maxResults={max_results}&startAt={start_at}"
        )
        response = requests.get(url, auth=auth, headers=headers, timeout=30)

        if not response.ok:
            raise JiraSyncError(
                f"Jira API returned {response.status_code}: {response.text[:200]}"
            )

        data = response.json()
        issues = data.get("issues", [])
        all_issues.extend(issues)

        total = data.get("total", 0)
        start_at += len(issues)
        if start_at >= total or not issues:
            break

    return all_issues


def _build_jira_key_index(existing_ids: Set[str], issues_dir: Path) -> Dict[str, str]:
    """Build a map from jira_key → kanbus identifier."""
    index: Dict[str, str] = {}
    for identifier in existing_ids:
        path = issue_path_for_identifier(issues_dir, identifier)
        try:
            issue = read_issue_from_file(path)
            jira_key = issue.custom.get("jira_key")
            if isinstance(jira_key, str):
                index[jira_key] = identifier
        except Exception:
            pass
    return index


def _jira_issue_key(issue: Dict[str, Any]) -> str:
    return issue.get("key", "")


def _jira_issue_summary(issue: Dict[str, Any]) -> str:
    return issue.get("fields", {}).get("summary", "Untitled")


def _map_jira_to_kanbus(
    jira_issue: Dict[str, Any],
    jira_config: JiraConfiguration,
    jira_key_to_kanbus_id: Dict[str, str],
) -> IssueData:
    """Map a Jira issue dict to a Kanbus IssueData."""
    fields = jira_issue.get("fields", {})
    jira_key = _jira_issue_key(jira_issue)

    title = fields.get("summary", "Untitled")
    description = _extract_adf_text(fields.get("description"))

    jira_type = (fields.get("issuetype") or {}).get("name", "Task")
    issue_type = jira_config.type_mappings.get(jira_type, jira_type.lower())

    jira_status = (fields.get("status") or {}).get("name", "open")
    status = _map_jira_status(jira_status)

    jira_priority = (fields.get("priority") or {}).get("name", "Medium")
    priority = _map_jira_priority(jira_priority)

    assignee_field = fields.get("assignee") or {}
    assignee: Optional[str] = assignee_field.get("displayName")

    reporter_field = fields.get("reporter") or {}
    creator: Optional[str] = reporter_field.get("displayName")

    parent_jira_key = (fields.get("parent") or {}).get("key")
    parent: Optional[str] = (
        jira_key_to_kanbus_id.get(parent_jira_key) if parent_jira_key else None
    )

    labels: List[str] = fields.get("labels") or []

    comments = _extract_comments(fields.get("comment") or {})

    created_at = _parse_jira_datetime(fields.get("created", "")) or datetime.now(
        timezone.utc
    )
    updated_at = _parse_jira_datetime(fields.get("updated", "")) or datetime.now(
        timezone.utc
    )
    closed_at = _parse_jira_datetime(fields.get("resolutiondate") or "")

    custom: Dict[str, object] = {"jira_key": jira_key}

    return IssueData.model_validate(
        {
            "id": "__placeholder__",  # filled in by caller
            "title": title,
            "description": description,
            "type": issue_type,
            "status": status,
            "priority": priority,
            "assignee": assignee,
            "creator": creator,
            "parent": parent,
            "labels": labels,
            "dependencies": [],
            "comments": [
                {
                    "id": c.id,
                    "author": c.author,
                    "text": c.text,
                    "created_at": c.created_at.isoformat(),
                }
                for c in comments
            ],
            "created_at": created_at.isoformat(),
            "updated_at": updated_at.isoformat(),
            "closed_at": closed_at.isoformat() if closed_at else None,
            "custom": custom,
        }
    )


def _extract_adf_text(value: Any) -> str:
    """Extract plain text from Atlassian Document Format or plain string."""
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        return _extract_adf_content(value)
    return ""


def _extract_adf_content(node: Dict[str, Any]) -> str:
    parts: List[str] = []
    for child in node.get("content", []):
        node_type = child.get("type", "")
        if node_type == "text":
            parts.append(child.get("text", ""))
        elif node_type == "hardBreak":
            parts.append("\n")
        else:
            text = _extract_adf_content(child)
            if text:
                parts.append(text)
    return " ".join(p for p in parts if p)


def _extract_comments(comment_field: Dict[str, Any]) -> List[IssueComment]:
    result = []
    for c in comment_field.get("comments", []):
        author = (c.get("author") or {}).get("displayName", "Unknown")
        text = _extract_adf_text(c.get("body"))
        created_at = _parse_jira_datetime(c.get("created", "")) or datetime.now(
            timezone.utc
        )
        result.append(
            IssueComment(
                id=str(c["id"]) if "id" in c else None,
                author=author,
                text=text or "(empty)",
                created_at=created_at,
            )
        )
    return result


def _map_jira_status(jira_status: str) -> str:
    mapping = {
        "to do": "open",
        "open": "open",
        "new": "open",
        "backlog": "open",
        "in progress": "in_progress",
        "in review": "in_progress",
        "in development": "in_progress",
        "done": "closed",
        "closed": "closed",
        "resolved": "closed",
        "complete": "closed",
        "completed": "closed",
        "blocked": "blocked",
        "impediment": "blocked",
    }
    return mapping.get(jira_status.lower(), "open")


def _map_jira_priority(jira_priority: str) -> int:
    mapping = {
        "highest": 0,
        "critical": 0,
        "blocker": 0,
        "high": 1,
        "medium": 2,
        "normal": 2,
        "low": 3,
        "lowest": 4,
        "trivial": 4,
        "minor": 4,
    }
    return mapping.get(jira_priority.lower(), 2)


def _parse_jira_datetime(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    except ValueError:
        return None


def issue_path_for_identifier(issues_dir: Path, identifier: str) -> Path:
    """Resolve an issue file path by identifier."""
    return issues_dir / f"{identifier}.json"



"""Issue comment management."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from uuid import uuid4

from kanbus.issue_files import write_issue_to_file
from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
from kanbus.models import IssueComment, IssueData
from kanbus.event_history import (
    comment_payload,
    comment_updated_payload,
    create_event,
    events_dir_for_issue_path,
    now_timestamp,
    write_events_batch,
)
from kanbus.users import get_current_user


class IssueCommentError(RuntimeError):
    """Raised when issue comment creation fails."""


@dataclass(frozen=True)
class IssueCommentResult:
    """Result of adding a comment to an issue."""

    issue: IssueData
    comment: IssueComment


def _generate_comment_id() -> str:
    return str(uuid4())


def _ensure_comment_ids(issue: IssueData) -> tuple[IssueData, bool]:
    changed = False
    comments = []
    for comment in issue.comments:
        if not comment.id:
            changed = True
            comments.append(
                IssueComment(
                    id=_generate_comment_id(),
                    author=comment.author,
                    text=comment.text,
                    created_at=comment.created_at,
                )
            )
        else:
            comments.append(comment)
    if not changed:
        return issue, False
    updated = issue.model_copy(update={"comments": comments})
    return updated, True


def _normalize_prefix(prefix: str) -> str:
    normalized = prefix.strip().lower()
    if not normalized:
        raise IssueCommentError("comment id is required")
    return normalized


def _find_comment_index(issue: IssueData, prefix: str) -> int:
    normalized = _normalize_prefix(prefix)
    matches: list[int] = []
    for index, comment in enumerate(issue.comments):
        if comment.id and comment.id.lower().startswith(normalized):
            matches.append(index)
    if not matches:
        raise IssueCommentError("comment not found")
    if len(matches) > 1:
        ids = ", ".join(
            (issue.comments[index].id or "")[:6]
            for index in matches
            if issue.comments[index].id
        )
        raise IssueCommentError(f"comment id prefix is ambiguous; matches: {ids}")
    return matches[0]


def add_comment(
    root: Path, identifier: str, author: str, text: str
) -> IssueCommentResult:
    """Add a comment to an issue.

    :param root: Repository root path.
    :type root: Path
    :param identifier: Issue identifier.
    :type identifier: str
    :param author: Comment author.
    :type author: str
    :param text: Comment text.
    :type text: str
    :return: Comment result including the updated issue.
    :rtype: IssueCommentResult
    :raises IssueCommentError: If the issue cannot be found or updated.
    """
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueCommentError(str(error)) from error

    timestamp = datetime.now(timezone.utc)
    base_issue, _ = _ensure_comment_ids(lookup.issue)
    comment = IssueComment(
        id=_generate_comment_id(),
        author=author,
        text=text,
        created_at=timestamp,
    )
    comments = [*base_issue.comments, comment]
    updated = lookup.issue.model_copy(
        update={"comments": comments, "updated_at": timestamp}
    )
    write_issue_to_file(updated, lookup.issue_path)
    comment_id = comment.id
    if not comment_id:
        raise IssueCommentError("comment id is required")
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=updated.identifier,
        event_type="comment_added",
        actor_id=actor_id,
        payload=comment_payload(comment_id, comment.author),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(lookup.project_dir, lookup.issue_path)
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(lookup.issue, lookup.issue_path)
        raise IssueCommentError(str(error)) from error
    return IssueCommentResult(issue=updated, comment=comment)


def ensure_issue_comment_ids(root: Path, identifier: str) -> IssueData:
    """Ensure comment ids are set for an issue and persist any changes."""
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueCommentError(str(error)) from error
    updated, changed = _ensure_comment_ids(lookup.issue)
    if changed:
        write_issue_to_file(updated, lookup.issue_path)
    return updated


def update_comment(
    root: Path, identifier: str, comment_id: str, text: str
) -> IssueData:
    """Update a comment by id prefix."""
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueCommentError(str(error)) from error
    issue, _ = _ensure_comment_ids(lookup.issue)
    index = _find_comment_index(issue, comment_id)
    comments = list(issue.comments)
    existing_comment = comments[index]
    updated_comment = comments[index].model_copy(update={"text": text})
    comments[index] = updated_comment
    updated = issue.model_copy(
        update={"comments": comments, "updated_at": datetime.now(timezone.utc)}
    )
    write_issue_to_file(updated, lookup.issue_path)
    if not existing_comment.id:
        raise IssueCommentError("comment id is required")
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=updated.identifier,
        event_type="comment_updated",
        actor_id=actor_id,
        payload=comment_updated_payload(existing_comment.id, existing_comment.author),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(lookup.project_dir, lookup.issue_path)
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(lookup.issue, lookup.issue_path)
        raise IssueCommentError(str(error)) from error
    return updated


def delete_comment(root: Path, identifier: str, comment_id: str) -> IssueData:
    """Delete a comment by id prefix."""
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueCommentError(str(error)) from error
    issue, _ = _ensure_comment_ids(lookup.issue)
    index = _find_comment_index(issue, comment_id)
    comments = list(issue.comments)
    removed_comment = comments.pop(index)
    updated = issue.model_copy(
        update={"comments": comments, "updated_at": datetime.now(timezone.utc)}
    )
    write_issue_to_file(updated, lookup.issue_path)
    if not removed_comment.id:
        raise IssueCommentError("comment id is required")
    occurred_at = now_timestamp()
    actor_id = get_current_user()
    event = create_event(
        issue_id=updated.identifier,
        event_type="comment_deleted",
        actor_id=actor_id,
        payload=comment_payload(removed_comment.id, removed_comment.author),
        occurred_at=occurred_at,
    )
    events_dir = events_dir_for_issue_path(lookup.project_dir, lookup.issue_path)
    try:
        write_events_batch(events_dir, [event])
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(lookup.issue, lookup.issue_path)
        raise IssueCommentError(str(error)) from error
    return updated



"""Query utilities for issue listing."""

from __future__ import annotations

from typing import Iterable, List

from kanbus.models import IssueData


class QueryError(RuntimeError):
    """Raised when query parameters are invalid."""


def filter_issues(
    issues: Iterable[IssueData],
    status: str | None,
    issue_type: str | None,
    assignee: str | None,
    label: str | None,
) -> List[IssueData]:
    """Filter issues by common fields.

    :param issues: Issues to filter.
    :type issues: Iterable[IssueData]
    :param status: Status filter.
    :type status: str | None
    :param issue_type: Type filter.
    :type issue_type: str | None
    :param assignee: Assignee filter.
    :type assignee: str | None
    :param label: Label filter.
    :type label: str | None
    :return: Filtered issues.
    :rtype: List[IssueData]
    """
    result = list(issues)
    if status:
        result = [issue for issue in result if issue.status == status]
    if issue_type:
        result = [issue for issue in result if issue.issue_type == issue_type]
    if assignee:
        result = [issue for issue in result if issue.assignee == assignee]
    if label:
        result = [issue for issue in result if label in issue.labels]
    return result


def sort_issues(issues: Iterable[IssueData], sort_key: str | None) -> List[IssueData]:
    """Sort issues by a supported key.

    :param issues: Issues to sort.
    :type issues: Iterable[IssueData]
    :param sort_key: Sort key name.
    :type sort_key: str | None
    :return: Sorted issues.
    :rtype: List[IssueData]
    :raises QueryError: If the sort key is unsupported.
    """
    result = list(issues)
    if sort_key is None:
        return result
    if sort_key == "priority":
        return sorted(result, key=lambda issue: issue.priority)
    raise QueryError("invalid sort key")


def search_issues(issues: Iterable[IssueData], term: str | None) -> List[IssueData]:
    """Search issues by title and description.

    :param issues: Issues to search.
    :type issues: Iterable[IssueData]
    :param term: Search term.
    :type term: str | None
    :return: Matching issues.
    :rtype: List[IssueData]
    """
    if not term:
        return list(issues)
    lowered = term.lower()
    matches: List[IssueData] = []
    seen: set[str] = set()
    for issue in issues:
        if lowered in issue.title.lower() or lowered in issue.description.lower():
            if issue.identifier not in seen:
                matches.append(issue)
                seen.add(issue.identifier)
            continue
        for comment in issue.comments:
            if lowered in comment.text.lower():
                if issue.identifier not in seen:
                    matches.append(issue)
                    seen.add(issue.identifier)
                break
    return matches



"""Kanbus CLI entry point."""

from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import Optional

import click

from kanbus import __version__
from kanbus.file_io import (
    InitializationError,
    detect_repairable_project_issues,
    ensure_git_repository,
    initialize_project,
    repair_project_structure,
)
from kanbus.content_validation import ContentValidationError, validate_code_blocks
from kanbus.rich_text_signals import apply_text_quality_signals, emit_signals
from kanbus.issue_creation import IssueCreationError, create_issue
from kanbus.issue_close import IssueCloseError, close_issue
from kanbus.issue_comment import IssueCommentError, add_comment
from kanbus.issue_delete import IssueDeleteError, delete_issue
from kanbus.beads_write import (
    BeadsDeleteError,
    BeadsWriteError,
    create_beads_issue,
    delete_beads_issue,
    update_beads_issue,
)
from kanbus.issue_display import format_issue_for_display
from kanbus.models import IssueData
from kanbus.ids import format_issue_key
from kanbus.issue_line import compute_widths, format_issue_line
from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
from kanbus.issue_update import IssueUpdateError, update_issue
from kanbus.issue_transfer import IssueTransferError, localize_issue, promote_issue
from kanbus.issue_listing import IssueListingError, list_issues
from kanbus.queries import QueryError
from kanbus.daemon_client import DaemonClientError, request_shutdown, request_status
from kanbus.users import get_current_user
from kanbus.migration import MigrationError, load_beads_issue, migrate_from_beads
from kanbus.doctor import DoctorError, run_doctor
from kanbus.maintenance import (
    ProjectStatsError,
    ProjectValidationError,
    collect_project_stats,
    validate_project,
)
from kanbus.dependencies import (
    DependencyError,
    add_dependency,
    list_ready_issues,
    remove_dependency,
)
from kanbus.dependency_tree import (
    DependencyTreeError,
    build_dependency_tree,
    render_dependency_tree,
)
from kanbus.wiki import WikiError, WikiRenderRequest, render_wiki_page
from kanbus.console_snapshot import ConsoleSnapshotError, build_console_snapshot
from kanbus.console_ui_state import fetch_console_ui_state
from kanbus.notification_publisher import publish_notification
from kanbus.project import ProjectMarkerError, get_configuration_path
from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.agents_management import _ensure_project_guard_files, ensure_agents_file


def _resolve_beads_mode(context: click.Context, beads_mode: bool) -> tuple[bool, bool]:
    source = context.get_parameter_source("beads_mode")
    if source == click.core.ParameterSource.COMMANDLINE and beads_mode:
        return True, True
    try:
        configuration = load_project_configuration(get_configuration_path(Path.cwd()))
    except ProjectMarkerError:
        return False, False
    except ConfigurationError as error:
        raise click.ClickException(str(error)) from error
    return configuration.beads_compatibility, False


@click.group()
@click.version_option(__version__, prog_name="kanbus")
@click.option("--beads", "beads_mode", is_flag=True, default=False)
@click.pass_context
def cli(context: click.Context, beads_mode: bool) -> None:
    """Kanbus issue tracker CLI.

    \b
    Quick start:
      kbs list                           list all issues
      kbs create "Fix bug" --type bug    create an issue
      kbs update <id> --status done      update an issue
      kbs comment <id> "Note"            add a comment
      kbs close <id>                     close an issue

    \b
    Issue types:  initiative > epic > story / task / bug > sub-task
    Statuses:     open  in_progress  blocked  done  closed
    Priorities:   0=critical  1=high  2=medium(default)  3=low  4=trivial
    """
    resolved, forced = _resolve_beads_mode(context, beads_mode)
    context.obj = {"beads_mode": resolved, "beads_mode_forced": forced}
    _maybe_prompt_project_repair(context)


def _should_check_project_structure(context: click.Context) -> bool:
    if context.invoked_subcommand is None:
        return False
    return context.invoked_subcommand not in {"init", "setup", "repair"}


def _maybe_prompt_project_repair(context: click.Context) -> None:
    if not _should_check_project_structure(context):
        return
    root = Path.cwd()
    try:
        plan = detect_repairable_project_issues(root, allow_uninitialized=True)
    except PermissionError:
        return
    if plan is None:
        return
    if not sys.stdin.isatty() or not sys.stdout.isatty():
        return
    missing = []
    if plan.missing_project_dir:
        missing.append("project/")
    if plan.missing_issues_dir:
        missing.append("project/issues")
    if plan.missing_events_dir:
        missing.append("project/events")
    prompt = (
        f"Project structure incomplete (missing: {', '.join(missing)}). Repair now?"
    )
    if click.confirm(prompt, default=False):
        repair_project_structure(root, plan)
        click.echo("Project structure repaired.", err=True)


@cli.group("setup")
def setup() -> None:
    """Setup utilities for Kanbus."""


@setup.command("agents")
@click.option("--force", is_flag=True, default=False)
def setup_agents(force: bool) -> None:
    """Ensure AGENTS.md contains Kanbus instructions.

    :param force: Overwrite existing Kanbus section without prompting.
    :type force: bool
    """
    root = Path.cwd()
    ensure_agents_file(root, force)
    _ensure_project_guard_files(root)


@cli.command("init")
@click.option("--local", "create_local", is_flag=True, default=False)
def init(create_local: bool) -> None:
    """Initialize a Kanbus project in the current repository.

    :param create_local: Whether to create a project-local directory.
    :type create_local: bool
    """
    root = Path.cwd()
    try:
        ensure_git_repository(root)
        initialize_project(root, create_local)
    except InitializationError as error:
        raise click.ClickException(str(error)) from error
    _maybe_run_setup_agents(root)


def _maybe_run_setup_agents(root: Path) -> None:
    if not sys.stdin.isatty() or not sys.stdout.isatty():
        return
    if click.confirm('Run "kanbus setup agents" now?', default=False):
        ensure_agents_file(root, force=False)


@cli.command("repair")
@click.option("--yes", is_flag=True, default=False)
def repair(yes: bool) -> None:
    """Repair a broken project structure."""
    root = Path.cwd()
    try:
        plan = detect_repairable_project_issues(root, allow_uninitialized=False)
    except ProjectMarkerError as error:
        raise click.ClickException(str(error)) from error
    except ConfigurationError as error:
        raise click.ClickException(str(error)) from error

    if plan is None:
        click.echo("Project structure is already healthy.")
        return

    if not yes:
        if not sys.stdin.isatty() or not sys.stdout.isatty():
            raise click.ClickException(
                "project structure requires repair (re-run with --yes)"
            )
        missing = []
        if plan.missing_project_dir:
            missing.append("project/")
        if plan.missing_issues_dir:
            missing.append("project/issues")
        if plan.missing_events_dir:
            missing.append("project/events")
        prompt = (
            f"Project structure incomplete (missing: {', '.join(missing)}). Repair now?"
        )
        if not click.confirm(prompt, default=False):
            click.echo("Repair cancelled.")
            return

    repair_project_structure(root, plan)
    click.echo("Project structure repaired.")


@cli.command("create")
@click.argument("title", nargs=-1)
@click.option("--type", "issue_type")
@click.option("--priority", type=int)
@click.option("--assignee")
@click.option("--parent")
@click.option("--label", "labels", multiple=True)
@click.option("--description", default="")
@click.option("--local", "local_issue", is_flag=True, default=False)
@click.option("--no-validate", "no_validate", is_flag=True, default=False)
@click.pass_context
def create(
    context: click.Context,
    title: tuple[str, ...],
    issue_type: str | None,
    priority: int | None,
    assignee: str | None,
    parent: str | None,
    labels: tuple[str, ...],
    description: str,
    local_issue: bool,
    no_validate: bool,
) -> None:
    """Create a new issue in the current project.

    \b
    Examples:
      kbs create "Plan the roadmap" --type initiative
      kbs create "Release v1" --type epic --parent <initiative-id>
      kbs create "Implement login" --type task --parent <epic-id>
      kbs create "Fix crash on launch" --type bug --priority 0 --parent <epic-id>

    :param title: Issue title words.
    :type title: tuple[str, ...]
    :param issue_type: Issue type override.
    :type issue_type: str | None
    :param priority: Issue priority override.
    :type priority: int | None
    :param assignee: Issue assignee.
    :type assignee: str | None
    :param parent: Parent issue identifier.
    :type parent: str | None
    :param labels: Issue labels.
    :type labels: tuple[str, ...]
    :param description: Issue description.
    :type description: str
    :param local_issue: Whether to create the issue in project-local.
    :type local_issue: bool
    """
    title_text = " ".join(title).strip()
    description_text = description.strip()
    if not title_text:
        raise click.ClickException("title is required")

    quality_result = None
    if description_text:
        quality_result = apply_text_quality_signals(description_text)
        description_text = quality_result.text

    if not no_validate and description_text:
        try:
            validate_code_blocks(description_text)
        except ContentValidationError as error:
            raise click.ClickException(str(error)) from error

    root = Path.cwd()
    beads_mode = bool(context.obj.get("beads_mode")) if context.obj else False
    if beads_mode:
        if local_issue:
            raise click.ClickException("beads mode does not support local issues")
        try:
            issue = create_beads_issue(
                root=root,
                title=title_text,
                issue_type=issue_type,
                priority=priority,
                assignee=assignee,
                parent=parent,
                description=description_text,
            )
        except BeadsWriteError as error:
            raise click.ClickException(str(error)) from error
        click.echo(
            format_issue_for_display(
                issue,
                configuration=None,
                project_context=False,
            )
        )
        if quality_result:
            emit_signals(quality_result, "description", issue_id=issue.identifier)
        return

    try:
        result = create_issue(
            root=root,
            title=title_text,
            issue_type=issue_type,
            priority=priority,
            assignee=assignee,
            parent=parent,
            labels=labels,
            description=description_text,
            local=local_issue,
            validate=not no_validate,
        )
    except IssueCreationError as error:
        raise click.ClickException(str(error)) from error

    click.echo(
        format_issue_for_display(
            result.issue,
            configuration=result.configuration,
            project_context=False,
        )
    )
    if quality_result:
        emit_signals(quality_result, "description", issue_id=result.issue.identifier)


@cli.command("show")
@click.argument("identifier")
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def show(context: click.Context, identifier: str, as_json: bool) -> None:
    """Show details for an issue.

    :param identifier: Issue identifier.
    :type identifier: str
    :param as_json: Emit JSON output when set.
    :type as_json: bool
    """
    root = Path.cwd()
    beads_mode = bool(context.obj.get("beads_mode")) if context.obj else False

    # Check if beads_compatibility is enabled in config
    if not beads_mode:
        try:
            config = load_project_configuration(get_configuration_path(root))
            if config.beads_compatibility:
                beads_mode = True
        except (ConfigurationError, ProjectMarkerError):
            pass

    if beads_mode:
        try:
            issue = load_beads_issue(root, identifier)
        except MigrationError as error:
            raise click.ClickException(str(error)) from error
        configuration = None
    else:
        try:
            lookup = load_issue_from_project(root, identifier)
        except IssueLookupError as error:
            raise click.ClickException(str(error)) from error
        issue = lookup.issue
        configuration = load_project_configuration(get_configuration_path(root))

    if as_json:
        payload = issue.model_dump(by_alias=True, mode="json")
        click.echo(json.dumps(payload, indent=2, sort_keys=False))
        return

    click.echo(
        format_issue_for_display(
            issue,
            configuration=configuration,
            project_context=False,
        )
    )


@cli.command("update")
@click.argument("identifier")
@click.option("--title")
@click.option("--description")
@click.option("--status")
@click.option("--priority", type=int)
@click.option("--assignee")
@click.option("--parent")
@click.option("--add-label", "add_labels", multiple=True)
@click.option("--remove-label", "remove_labels", multiple=True)
@click.option("--set-labels", "set_labels")
@click.option("--claim", is_flag=True, default=False)
@click.option("--no-validate", "no_validate", is_flag=True, default=False)
def update(
    identifier: str,
    title: str | None,
    description: str | None,
    status: str | None,
    priority: int | None,
    assignee: str | None,
    parent: str | None,
    add_labels: tuple[str, ...],
    remove_labels: tuple[str, ...],
    set_labels: str | None,
    claim: bool,
    no_validate: bool,
) -> None:
    """Update an existing issue.

    :param identifier: Issue identifier.
    :type identifier: str
    :param title: Updated title.
    :type title: str | None
    :param description: Updated description.
    :type description: str | None
    :param status: Updated status.
    :type status: str | None
    :param priority: Updated priority.
    :type priority: int | None
    :param assignee: Updated assignee.
    :type assignee: str | None
    :param parent: Updated parent identifier.
    :type parent: str | None
    :param claim: Whether to claim the issue.
    :type claim: bool
    """
    root = Path.cwd()
    beads_mode = False
    if click.get_current_context().obj:
        beads_mode = bool(click.get_current_context().obj.get("beads_mode"))

    # Check if beads_compatibility is enabled in config
    if not beads_mode:
        try:
            config = load_project_configuration(get_configuration_path(root))
            if config.beads_compatibility:
                beads_mode = True
        except (ConfigurationError, ProjectMarkerError):
            pass

    update_quality_result = None
    if description:
        description_text_stripped = description.strip()
        if description_text_stripped:
            update_quality_result = apply_text_quality_signals(
                description_text_stripped
            )
            description = update_quality_result.text

    if not no_validate and description:
        try:
            validate_code_blocks(description)
        except ContentValidationError as error:
            raise click.ClickException(str(error)) from error

    # Parse set_labels if provided
    parsed_set_labels = None
    if set_labels:
        parsed_set_labels = [label.strip() for label in set_labels.split(",")]

    if beads_mode:
        if parent is not None:
            raise click.ClickException("parent update not supported in beads mode")
        try:
            update_beads_issue(
                root,
                identifier,
                status=status,
                title=title.strip() if title else None,
                description=description.strip() if description else None,
                priority=priority,
                assignee=assignee,
                add_labels=list(add_labels) if add_labels else None,
                remove_labels=list(remove_labels) if remove_labels else None,
                set_labels=parsed_set_labels,
            )
        except BeadsWriteError as error:
            raise click.ClickException(str(error)) from error
        formatted_identifier = format_issue_key(identifier, project_context=False)
        click.echo(f"Updated {formatted_identifier}")
        if update_quality_result:
            emit_signals(
                update_quality_result,
                "description",
                issue_id=identifier,
                is_update=True,
            )
        return

    # Regular Kanbus mode
    try:
        final_assignee = assignee or (get_current_user() if claim else None)
        update_issue(
            root=root,
            identifier=identifier,
            title=title.strip() if title else None,
            description=description.strip() if description else None,
            status=status,
            assignee=final_assignee,
            claim=claim,
            validate=not no_validate,
            priority=priority,
            add_labels=list(add_labels) if add_labels else None,
            remove_labels=list(remove_labels) if remove_labels else None,
            set_labels=parsed_set_labels,
            parent=parent,
        )
    except IssueUpdateError as error:
        raise click.ClickException(str(error)) from error

    formatted_identifier = format_issue_key(identifier, project_context=False)
    click.echo(f"Updated {formatted_identifier}")
    if update_quality_result:
        emit_signals(
            update_quality_result, "description", issue_id=identifier, is_update=True
        )


@cli.command("close")
@click.argument("identifier")
def close(identifier: str) -> None:
    """Close an issue.

    :param identifier: Issue identifier.
    :type identifier: str
    """
    root = Path.cwd()
    try:
        close_issue(root, identifier)
    except IssueCloseError as error:
        raise click.ClickException(str(error)) from error
    formatted_identifier = format_issue_key(identifier, project_context=False)
    click.echo(f"Closed {formatted_identifier}")


@cli.command("delete")
@click.argument("identifier")
def delete(identifier: str) -> None:
    """Delete an issue.

    :param identifier: Issue identifier.
    :type identifier: str
    """
    root = Path.cwd()
    beads_mode = bool(click.get_current_context().obj.get("beads_mode"))

    # Check if beads_compatibility is enabled in config
    if not beads_mode:
        try:
            config = load_project_configuration(get_configuration_path(root))
            if config.beads_compatibility:
                beads_mode = True
        except (ConfigurationError, ProjectMarkerError):
            pass

    if beads_mode:
        try:
            delete_beads_issue(root, identifier)
        except BeadsDeleteError as error:
            raise click.ClickException(str(error)) from error
    else:
        try:
            delete_issue(root, identifier)
        except IssueDeleteError as error:
            raise click.ClickException(str(error)) from error
    formatted_identifier = format_issue_key(identifier, project_context=False)
    click.echo(f"Deleted {formatted_identifier}")


@cli.command("promote")
@click.argument("identifier")
def promote(identifier: str) -> None:
    """Promote a local issue to the shared project.

    :param identifier: Issue identifier.
    :type identifier: str
    """
    root = Path.cwd()
    try:
        promote_issue(root, identifier)
    except IssueTransferError as error:
        raise click.ClickException(str(error)) from error


@cli.command("localize")
@click.argument("identifier")
def localize(identifier: str) -> None:
    """Move a shared issue into project-local.

    :param identifier: Issue identifier.
    :type identifier: str
    """
    root = Path.cwd()
    try:
        localize_issue(root, identifier)
    except IssueTransferError as error:
        raise click.ClickException(str(error)) from error


@cli.command("comment")
@click.argument("identifier")
@click.argument("text", required=False)
@click.option("--body-file", type=click.File("r"), default=None)
@click.option("--no-validate", "no_validate", is_flag=True, default=False)
@click.pass_context
def comment(
    context: click.Context,
    identifier: str,
    text: Optional[str],
    body_file: Optional[click.File],
    no_validate: bool = False,
) -> None:
    """Add a comment to an issue.

    :param context: Click context.
    :type context: click.Context
    :param identifier: Issue identifier.
    :type identifier: str
    :param text: Comment text (or use --body-file for multi-line).
    :type text: Optional[str]
    :param body_file: File to read comment text from (use '-' for stdin).
    :type body_file: Optional[click.File]
    :param no_validate: Bypass validation checks.
    :type no_validate: bool
    """
    root = Path.cwd()
    beads_mode = context.obj.get("beads_mode", False)

    # Check if beads_compatibility is enabled in config
    if not beads_mode:
        try:
            config = load_project_configuration(get_configuration_path(root))
            if config.beads_compatibility:
                beads_mode = True
        except (ConfigurationError, ProjectMarkerError):
            pass

    # Handle body-file input
    comment_text = text or ""
    if body_file is not None:
        comment_text = body_file.read()

    if not comment_text:
        raise click.ClickException("Comment text required")

    comment_quality_result = apply_text_quality_signals(comment_text)
    comment_text = comment_quality_result.text

    if not no_validate:
        try:
            validate_code_blocks(comment_text)
        except ContentValidationError as error:
            raise click.ClickException(str(error)) from error

    try:
        if beads_mode:
            from kanbus.beads_write import add_beads_comment, BeadsWriteError

            try:
                add_beads_comment(
                    root=root,
                    identifier=identifier,
                    author=get_current_user(),
                    text=comment_text,
                )
            except BeadsWriteError as error:
                raise click.ClickException(str(error)) from error
            emit_signals(comment_quality_result, "comment", issue_id=identifier)
        else:
            result_comment = add_comment(
                root=root,
                identifier=identifier,
                author=get_current_user(),
                text=comment_text,
            )
            emit_signals(
                comment_quality_result,
                "comment",
                issue_id=identifier,
                comment_id=result_comment.comment.id,
            )
    except IssueCommentError as error:
        raise click.ClickException(str(error)) from error


@cli.command("list")
@click.option("--status")
@click.option("--type", "issue_type")
@click.option("--assignee")
@click.option("--label")
@click.option("--sort")
@click.option("--search")
@click.option("--project", "projects", multiple=True, help="Filter by project label.")
@click.option("--no-local", is_flag=True, default=False)
@click.option("--local-only", is_flag=True, default=False)
@click.option(
    "--limit",
    type=int,
    default=50,
    show_default=True,
    help="Maximum issues to display (0 for no limit). Matches Beads default.",
)
@click.option(
    "--porcelain",
    is_flag=True,
    default=False,
    help="Plain, non-colorized output for machine parsing.",
)
@click.pass_context
def list_command(
    context: click.Context,
    status: str | None,
    issue_type: str | None,
    assignee: str | None,
    label: str | None,
    sort: str | None,
    search: str | None,
    projects: tuple[str, ...],
    no_local: bool,
    local_only: bool,
    limit: int,
    porcelain: bool,
) -> None:
    """List issues in the current project.

    \b
    Examples:
      kbs list
      kbs list --type epic
      kbs list --status open
      kbs list --type task --status in_progress
      kbs issues / kbs epics / kbs tasks / kbs bugs   shorthand aliases
    """
    root = Path.cwd()
    beads_mode = bool(context.obj.get("beads_mode")) if context.obj else False
    try:
        issues = list_issues(
            root,
            status=status,
            issue_type=issue_type,
            assignee=assignee,
            label=label,
            sort=sort,
            search=search,
            project_filter=list(projects),
            include_local=not no_local,
            local_only=local_only,
            beads_mode=beads_mode,
        )
    except (IssueListingError, QueryError) as error:
        raise click.ClickException(str(error)) from error

    if beads_mode:
        issues = sorted(
            issues,
            key=lambda issue: (
                issue.priority,
                -_issue_sort_timestamp(issue),
                issue.identifier,
            ),
        )
    if limit > 0:
        issues = issues[:limit]

    configuration = None
    if not beads_mode:
        try:
            configuration = load_project_configuration(get_configuration_path(root))
        except ProjectMarkerError:
            configuration = None
        except ConfigurationError as error:
            raise click.ClickException(str(error)) from error

    # In Beads mode, always show full IDs (project_context=False)
    # In regular mode, use project_context if all issues are from same project
    project_context = (
        False
        if beads_mode
        else not any(issue.custom.get("project_path") for issue in issues)
    )
    widths = (
        None if porcelain else compute_widths(issues, project_context=project_context)
    )
    for issue in issues:
        line = format_issue_line(
            issue,
            porcelain=porcelain,
            widths=widths,
            project_context=project_context,
            configuration=configuration,
        )
        click.echo(line)


def _issue_sort_timestamp(issue: IssueData) -> float:
    """Return a sortable UTC timestamp (seconds) for an issue."""

    timestamp = issue.closed_at or issue.updated_at or issue.created_at
    return timestamp.timestamp()


@cli.group("wiki")
def wiki() -> None:
    """Manage wiki pages."""


@wiki.command("render")
@click.argument("page")
def render_wiki(page: str) -> None:
    """Render a wiki page.

    :param page: Wiki page path.
    :type page: str
    """
    root = Path.cwd()
    request = WikiRenderRequest(root=root, page_path=Path(page))
    try:
        output = render_wiki_page(request)
    except WikiError as error:
        raise click.ClickException(str(error)) from error
    click.echo(output)


@cli.group("policy")
def policy() -> None:
    """Policy management commands."""


@policy.command("check")
@click.argument("identifier")
def policy_check(identifier: str) -> None:
    """Check policies against an issue.

    :param identifier: Issue identifier to check policies against.
    :type identifier: str
    """
    from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
    from kanbus.policy_loader import load_policies
    from kanbus.policy_evaluator import (
        evaluate_policies_with_options,
        PolicyEvaluationOptions,
    )
    from kanbus.policy_context import PolicyContext, PolicyOperation
    from kanbus.issue_listing import load_issues_from_directory
    from kanbus.config_loader import load_project_configuration
    from kanbus.project import get_configuration_path

    root = Path.cwd()
    try:
        lookup = load_issue_from_project(root, identifier)
        configuration = load_project_configuration(
            get_configuration_path(lookup.project_dir)
        )
        policies_dir = lookup.project_dir / "policies"

        if not policies_dir.is_dir():
            click.echo("No policies directory found")
            return

        policy_documents = load_policies(policies_dir)
        if not policy_documents:
            click.echo("No policy files found")
            return

        issues_dir = lookup.project_dir / "issues"
        all_issues = load_issues_from_directory(issues_dir)
        context = PolicyContext(
            current_issue=lookup.issue,
            proposed_issue=lookup.issue,
            transition=None,
            operation=PolicyOperation.UPDATE,
            project_configuration=configuration,
            all_issues=all_issues,
        )

        violations = evaluate_policies_with_options(
            context,
            policy_documents,
            PolicyEvaluationOptions(collect_all_violations=True),
        )

        if violations:
            error_msg = [f"Found {len(violations)} policy violation(s):"]
            for i, v in enumerate(violations):
                error_msg.append(f"\n{i + 1}. {v}")
            raise click.ClickException("\n".join(error_msg))

        click.echo(f"All policies passed for {identifier}")
    except IssueLookupError as error:
        raise click.ClickException(str(error)) from error
    except Exception as error:
        raise click.ClickException(str(error)) from error


@policy.command("list")
def policy_list() -> None:
    """List all loaded policy files."""
    from kanbus.project import load_project_directory
    from kanbus.policy_loader import load_policies

    root = Path.cwd()
    try:
        project_dir = load_project_directory(root)
        policies_dir = project_dir / "policies"

        if not policies_dir.is_dir():
            click.echo("No policies directory found")
            return

        policy_documents = load_policies(policies_dir)
        if not policy_documents:
            click.echo("No policy files found")
            return

        for filename, document in policy_documents:
            click.echo(f"{filename}")
            if document.feature:
                click.echo(f"  Feature: {document.feature.name}")
                for child in document.feature.children:
                    if hasattr(child, "scenario") and child.scenario:
                        click.echo(f"    Scenario: {child.scenario.name}")
    except Exception as error:
        raise click.ClickException(str(error)) from error


@policy.command("steps")
@click.option("--category", help="Filter by category (given, when, then).")
@click.option("--search", help="Filter by search term.")
def policy_steps(category: str | None, search: str | None) -> None:
    """List available policy steps.

    :param category: Filter by category.
    :type category: str | None
    :param search: Filter by search term.
    :type search: str | None
    """
    from kanbus.policy_evaluator import _get_step_registry

    registry = _get_step_registry()
    output = []

    for step in registry.steps:
        if category:
            if category.lower() != step.category.value.lower():
                continue
        if search:
            search_lower = search.lower()
            if (
                search_lower not in step.description.lower()
                and search_lower not in step.usage_pattern.lower()
            ):
                continue
        output.append(
            f"{step.category.value} - {step.description}\n  Pattern: {step.usage_pattern}"
        )

    if not output:
        click.echo("No matching steps found")
    else:
        click.echo("\n".join(output))


@policy.command("validate")
def policy_validate() -> None:
    """Validate all policy files for syntax errors."""
    from kanbus.project import load_project_directory
    from kanbus.policy_loader import load_policies

    root = Path.cwd()
    try:
        project_dir = load_project_directory(root)
        policies_dir = project_dir / "policies"

        if not policies_dir.is_dir():
            click.echo("No policies directory found")
            return

        policy_documents = load_policies(policies_dir)
        if not policy_documents:
            click.echo("No policy files found")
            return

        click.echo(f"All {len(policy_documents)} policy files are valid")
    except Exception as error:
        raise click.ClickException(str(error)) from error


@cli.group("console")
def console() -> None:
    """Console-related utilities."""


@console.command("snapshot")
def console_snapshot() -> None:
    """Emit a JSON snapshot for the console."""
    root = Path.cwd()
    try:
        snapshot = build_console_snapshot(root)
    except ConsoleSnapshotError as error:
        raise click.ClickException(str(error)) from error
    payload = json.dumps(snapshot, indent=2, sort_keys=False)
    click.echo(payload)


@console.command("focus")
@click.argument("identifier")
@click.option("--comment", default=None, help="Comment ID to scroll to.")
def console_focus(identifier: str, comment: Optional[str]) -> None:
    """Focus on an issue and its descendants in the console."""
    from kanbus.issue_lookup import IssueLookupError, load_issue_from_project

    root = Path.cwd()
    try:
        result = load_issue_from_project(root, identifier)
        issue_id = result.issue.identifier
    except IssueLookupError as error:
        raise click.ClickException(str(error)) from error
    event: dict = {
        "type": "issue_focused",
        "issue_id": issue_id,
    }
    if comment:
        event["comment_id"] = comment
    publish_notification(root, event)
    if comment:
        click.echo(f"Focused on issue {issue_id} (comment {comment})")
    else:
        click.echo(f"Focused on issue {issue_id}")


@console.command("unfocus")
def console_unfocus() -> None:
    """Clear the current focus filter in the console."""
    root = Path.cwd()
    publish_notification(
        root, {"type": "ui_control", "action": {"action": "clear_focus"}}
    )
    click.echo("Cleared focus filter")


@console.command("view")
@click.argument("mode", type=click.Choice(["initiatives", "epics", "issues"]))
def console_view(mode: str) -> None:
    """Switch the console to a different view mode."""
    root = Path.cwd()
    publish_notification(
        root,
        {"type": "ui_control", "action": {"action": "set_view_mode", "mode": mode}},
    )
    click.echo(f"Switched to {mode} view")


@console.command("search")
@click.argument("query", required=False, default=None)
@click.option("--clear", is_flag=True, help="Clear the active search query.")
def console_search(query: Optional[str], clear: bool) -> None:
    """Set or clear the search query in the console."""
    root = Path.cwd()
    if clear:
        search_query = ""
    elif query:
        search_query = query
    else:
        raise click.UsageError("Provide a query or use --clear.")
    publish_notification(
        root,
        {
            "type": "ui_control",
            "action": {"action": "set_search", "query": search_query},
        },
    )
    if clear or not search_query:
        click.echo("Cleared search query")
    else:
        click.echo(f"Set search query to: {search_query}")


@console.command("status")
def console_status() -> None:
    """Print a human-readable summary of the current console UI state."""
    root = Path.cwd()
    ui_state = fetch_console_ui_state(root)
    if ui_state is None:
        click.echo("Console server is not running.")
        return
    focused = ui_state.get("focused_issue_id") or "none"
    view = ui_state.get("view_mode") or "none"
    search = ui_state.get("search_query") or "none"
    click.echo(f"focus:  {focused}")
    click.echo(f"view:   {view}")
    click.echo(f"search: {search}")


@console.group("get")
def console_get() -> None:
    """Query a specific piece of console UI state."""


@console_get.command("focus")
def console_get_focus() -> None:
    """Print the currently focused issue ID, or 'none'."""
    root = Path.cwd()
    ui_state = fetch_console_ui_state(root)
    if ui_state is None:
        click.echo("Console server is not running.")
        return
    click.echo(ui_state.get("focused_issue_id") or "none")


@console_get.command("view")
def console_get_view() -> None:
    """Print the current view mode, or 'none'."""
    root = Path.cwd()
    ui_state = fetch_console_ui_state(root)
    if ui_state is None:
        click.echo("Console server is not running.")
        return
    click.echo(ui_state.get("view_mode") or "none")


@console_get.command("search")
def console_get_search() -> None:
    """Print the active search query, or 'none'."""
    root = Path.cwd()
    ui_state = fetch_console_ui_state(root)
    if ui_state is None:
        click.echo("Console server is not running.")
        return
    click.echo(ui_state.get("search_query") or "none")


@cli.command("validate")
def validate() -> None:
    """Validate project integrity."""
    root = Path.cwd()
    try:
        validate_project(root)
    except ProjectValidationError as error:
        raise click.ClickException(str(error)) from error


@cli.command("stats")
def stats() -> None:
    """Report project statistics."""
    root = Path.cwd()
    try:
        stats_result = collect_project_stats(root)
    except ProjectStatsError as error:
        raise click.ClickException(str(error)) from error

    lines = [
        f"total issues: {stats_result.total}",
        f"open issues: {stats_result.open_count}",
        f"closed issues: {stats_result.closed_count}",
    ]
    for issue_type in sorted(stats_result.type_counts):
        count = stats_result.type_counts[issue_type]
        lines.append(f"type: {issue_type}: {count}")
    click.echo("\n".join(lines))


@cli.command(
    "dep",
    context_settings={"ignore_unknown_options": True, "allow_interspersed_args": False},
)
@click.argument("args", nargs=-1, required=True)
@click.pass_context
def dep(context: click.Context, args: tuple[str, ...]) -> None:
    """Manage issue dependencies.

    Usage: kanbus dep <identifier> <blocked-by|relates-to> <target>
           kanbus dep <identifier> remove <blocked-by|relates-to> <target>
           kanbus dep tree <identifier> [--depth N] [--format FORMAT]
    """
    if len(args) < 1:
        raise click.ClickException("usage: kanbus dep <identifier> <type> <target>")

    # Check if this is a tree command - handle it separately since it needs options
    if args[0] == "tree":
        # Redirect to separate tree implementation
        if len(args) < 2:
            raise click.ClickException("tree requires an identifier")

        # For tree, we need to parse options differently
        # Just pass through to the tree handler with a simple parse
        tree_args = list(args[1:])
        tree_identifier = tree_args[0] if tree_args else None
        if not tree_identifier:
            raise click.ClickException("tree requires an identifier")

        # Simple option parsing
        depth = None
        output_format = "text"
        i = 1
        while i < len(tree_args):
            if tree_args[i] == "--depth" and i + 1 < len(tree_args):
                try:
                    depth = int(tree_args[i + 1])
                except ValueError:
                    raise click.ClickException("depth must be a number")
                i += 2
            elif tree_args[i] == "--format" and i + 1 < len(tree_args):
                output_format = tree_args[i + 1]
                i += 2
            else:
                i += 1

        root = Path.cwd()
        try:
            tree = build_dependency_tree(root, tree_identifier, depth)
            output = render_dependency_tree(tree, output_format)
        except DependencyTreeError as error:
            raise click.ClickException(str(error)) from error
        click.echo(output)
        return

    if len(args) < 2:
        raise click.ClickException("usage: kanbus dep <identifier> <type> <target>")

    identifier = args[0]

    # Check if this is a remove operation
    if len(args) > 1 and args[1] == "remove":
        if len(args) < 4:
            raise click.ClickException("dependency target is required")
        dep_type = args[2]
        target = args[3]
        is_remove = True
    else:
        if len(args) < 3:
            raise click.ClickException("dependency target is required")
        dep_type = args[1]
        target = args[2]
        is_remove = False

    root = Path.cwd()
    beads_mode = bool(context.obj.get("beads_mode")) if context.obj else False

    # Check if beads_compatibility is enabled in config
    if not beads_mode:
        try:
            config = load_project_configuration(get_configuration_path(root))
            if config.beads_compatibility:
                beads_mode = True
        except (ConfigurationError, ProjectMarkerError):
            pass

    if is_remove:
        if beads_mode:
            try:
                from kanbus.beads_write import remove_beads_dependency

                remove_beads_dependency(root, identifier, target, dep_type)
            except BeadsWriteError as error:
                raise click.ClickException(str(error)) from error
        else:
            try:
                remove_dependency(root, identifier, target, dep_type)
            except DependencyError as error:
                raise click.ClickException(str(error)) from error
    else:
        if beads_mode:
            try:
                from kanbus.beads_write import add_beads_dependency

                add_beads_dependency(root, identifier, target, dep_type)
            except BeadsWriteError as error:
                raise click.ClickException(str(error)) from error
        else:
            try:
                add_dependency(root, identifier, target, dep_type)
            except DependencyError as error:
                raise click.ClickException(str(error)) from error


@cli.command("ready")
@click.option("--no-local", is_flag=True, default=False)
@click.option("--local-only", is_flag=True, default=False)
@click.pass_context
def ready(context: click.Context, no_local: bool, local_only: bool) -> None:
    """List issues that are ready (not blocked)."""
    root = Path.cwd()
    beads_mode = bool(context.obj.get("beads_mode")) if context.obj else False
    try:
        issues = list_ready_issues(
            root,
            include_local=not no_local,
            local_only=local_only,
            beads_mode=beads_mode,
        )
    except DependencyError as error:
        raise click.ClickException(str(error)) from error
    for issue in issues:
        project_path = issue.custom.get("project_path")
        prefix = f"{project_path} " if project_path else ""
        click.echo(f"{prefix}{issue.identifier}")


@cli.command("doctor")
def doctor() -> None:
    """Run environment diagnostics for Kanbus."""
    root = Path.cwd()
    try:
        result = run_doctor(root)
    except DoctorError as error:
        raise click.ClickException(str(error)) from error
    click.echo(f"ok {result.project_dir}")


@cli.command("migrate")
def migrate() -> None:
    """Migrate Beads issues into Kanbus.

    :raises click.ClickException: If migration fails.
    """
    root = Path.cwd()
    try:
        result = migrate_from_beads(root)
    except MigrationError as error:
        raise click.ClickException(str(error)) from error
    click.echo(f"migrated {result.issue_count} issues")


@cli.command("daemon-status")
def daemon_status() -> None:
    """Report daemon status."""
    root = Path.cwd()
    try:
        result = request_status(root)
    except ProjectMarkerError as error:
        raise click.ClickException(_format_project_marker_error(error)) from error
    except DaemonClientError as error:
        raise click.ClickException(str(error)) from error
    click.echo(json.dumps(result, indent=2, sort_keys=False))


@cli.command("daemon-stop")
def daemon_stop() -> None:
    """Stop the daemon process."""
    root = Path.cwd()
    try:
        result = request_shutdown(root)
    except ProjectMarkerError as error:
        raise click.ClickException(_format_project_marker_error(error)) from error
    except DaemonClientError as error:
        raise click.ClickException(str(error)) from error
    click.echo(json.dumps(result, indent=2, sort_keys=False))


def _format_project_marker_error(error: ProjectMarkerError) -> str:
    message = str(error)
    if message.startswith("multiple projects found"):
        return (
            "multiple projects found. Run this command from a directory containing a "
            "single project/ folder."
        )
    if message == "project not initialized":
        return 'project not initialized. Run "kanbus init" to create a project/ folder.'
    return message


@cli.group()
def snyk() -> None:
    """Snyk vulnerability synchronization commands."""


@snyk.command(name="pull")
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Show what would be done without writing files.",
)
@click.option(
    "--min-severity",
    default=None,
    help="Override minimum severity (critical, high, medium, low).",
)
@click.option(
    "--org-id",
    default=None,
    help="Override Snyk org ID.",
)
@click.option(
    "--parent-epic",
    default=None,
    help="Override parent epic issue ID to attach bugs to.",
)
@click.pass_context
def snyk_pull(
    context: click.Context,
    dry_run: bool,
    min_severity: Optional[str],
    org_id: Optional[str],
    parent_epic: Optional[str],
) -> None:
    """Pull vulnerabilities from Snyk into Kanbus."""
    from kanbus.snyk_sync import SnykSyncError, pull_from_snyk

    root = Path.cwd()
    try:
        config_path = get_configuration_path(root)
        configuration = load_project_configuration(config_path)
    except ProjectMarkerError as error:
        raise click.ClickException(_format_project_marker_error(error)) from error
    except ConfigurationError as error:
        raise click.ClickException(str(error)) from error

    if configuration.snyk is None:
        raise click.ClickException("no snyk configuration in .kanbus.yml")

    snyk_config = configuration.snyk.model_copy(
        update={
            k: v
            for k, v in {
                "min_severity": min_severity,
                "org_id": org_id,
                "parent_epic": parent_epic,
            }.items()
            if v is not None
        }
    )

    if dry_run:
        click.echo("Dry run — no files will be written.\n")

    try:
        result = pull_from_snyk(root, snyk_config, configuration.project_key, dry_run)
    except SnykSyncError as error:
        raise click.ClickException(str(error)) from error

    click.echo(
        f"pulled {result.pulled} new, updated {result.updated} existing, "
        f"skipped {result.skipped} duplicates"
    )


@cli.group()
def jira() -> None:
    """Jira synchronization commands."""


@jira.command(name="pull")
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Show what would be done without writing files.",
)
@click.pass_context
def jira_pull(context: click.Context, dry_run: bool) -> None:
    """Pull issues from Jira into Kanbus."""
    from kanbus.jira_sync import JiraSyncError, pull_from_jira

    root = Path.cwd()
    try:
        config_path = get_configuration_path(root)
        configuration = load_project_configuration(config_path)
    except ProjectMarkerError as error:
        raise click.ClickException(_format_project_marker_error(error)) from error
    except ConfigurationError as error:
        raise click.ClickException(str(error)) from error

    if configuration.jira is None:
        raise click.ClickException("no jira configuration in .kanbus.yml")

    if configuration.jira.sync_direction not in ("pull", "both"):
        raise click.ClickException(
            "sync_direction must be 'pull' or 'both' to use jira pull"
        )

    if dry_run:
        click.echo("Dry run — no files will be written.\n")

    try:
        result = pull_from_jira(
            root, configuration.jira, configuration.project_key, dry_run
        )
    except JiraSyncError as error:
        raise click.ClickException(str(error)) from error

    click.echo(f"pulled {result.pulled} new, updated {result.updated} existing")


@cli.command("issues", hidden=True)
@click.pass_context
def issues_alias(context: click.Context) -> None:
    """Alias for: kbs list"""
    context.invoke(list_command)


@cli.command("epics", hidden=True)
@click.pass_context
def epics_alias(context: click.Context) -> None:
    """Alias for: kbs list --type epic"""
    context.invoke(list_command, issue_type="epic")


@cli.command("tasks", hidden=True)
@click.pass_context
def tasks_alias(context: click.Context) -> None:
    """Alias for: kbs list --type task"""
    context.invoke(list_command, issue_type="task")


@cli.command("stories", hidden=True)
@click.pass_context
def stories_alias(context: click.Context) -> None:
    """Alias for: kbs list --type story"""
    context.invoke(list_command, issue_type="story")


@cli.command("bugs", hidden=True)
@click.pass_context
def bugs_alias(context: click.Context) -> None:
    """Alias for: kbs list --type bug"""
    context.invoke(list_command, issue_type="bug")


if __name__ == "__main__":
    cli()



"""
Issue identifier generation.
"""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import Iterable, Optional, Set

from pydantic import BaseModel, Field


class IssueIdentifierRequest(BaseModel):
    """
    Request to generate a unique issue identifier.

    :param title: The issue title for uniqueness checks.
    :type title: str
    :param existing_ids: Set of existing IDs to avoid collisions.
    :type existing_ids: Set[str]
    :param prefix: ID prefix from configuration.
    :type prefix: str
    """

    title: str = Field(min_length=1)
    existing_ids: Set[str] = Field(default_factory=set)
    prefix: str = Field(default="kanbus", min_length=1)


@dataclass(frozen=True)
class IssueIdentifierResult:
    """Result of issue identifier generation."""

    identifier: str


_TEST_UUID_SEQUENCE: Optional[list[str]] = None


def set_test_uuid_sequence(sequence: Optional[Iterable[str]]) -> None:
    """
    Override UUID generation for deterministic tests.

    :param sequence: Sequence of UUID strings to use, or None to clear.
    :type sequence: Optional[Iterable[str]]
    """
    global _TEST_UUID_SEQUENCE
    _TEST_UUID_SEQUENCE = list(sequence) if sequence is not None else None


def _next_uuid_value() -> str:
    if _TEST_UUID_SEQUENCE:
        return _TEST_UUID_SEQUENCE.pop(0)
    return str(uuid.uuid4())


def format_issue_key(identifier: str, project_context: bool) -> str:
    """
    Produce a display-friendly issue key.

    :param identifier: Full issue identifier (may include project key and UUID).
    :type identifier: str
    :param project_context: Whether the display is within a project context.
    :type project_context: bool
    :return: Formatted key with optional project key and abbreviated hash.
    :rtype: str
    """
    if identifier.isdigit():
        return identifier

    key_part = ""
    remainder = identifier
    if "-" in identifier:
        parts = identifier.split("-", 1)
        if len(parts) == 2 and parts[0] and parts[1]:
            key_part, remainder = parts

    base = remainder
    suffix = ""
    if "." in remainder:
        base, suffix = remainder.split(".", 1)
        suffix = f".{suffix}"

    normalized = base.replace("-", "")
    truncated = normalized[:6] if normalized else normalized

    if project_context:
        return f"{truncated}{suffix}"

    if key_part:
        return f"{key_part}-{truncated}{suffix}"

    return f"{truncated}{suffix}"


def matches_issue_identifier(candidate: str, full_id: str) -> bool:
    """
    Check if a candidate identifier matches a full issue identifier.

    Accepts full identifiers, project-context short ids, and abbreviated prefixes.

    :param candidate: User-provided identifier value.
    :type candidate: str
    :param full_id: Full issue identifier from storage.
    :type full_id: str
    :return: True if the candidate matches the full identifier.
    :rtype: bool
    """
    if candidate == full_id:
        return True

    if candidate == format_issue_key(full_id, project_context=False):
        return True

    if candidate == format_issue_key(full_id, project_context=True):
        return True

    if len(candidate) >= len(full_id):
        return False

    return full_id.startswith(candidate)


def generate_issue_identifier(request: IssueIdentifierRequest) -> IssueIdentifierResult:
    """Generate a unique issue ID using a UUID.

    :param request: Validated request containing title and existing IDs.
    :type request: IssueIdentifierRequest
    :return: A unique ID string with format '{prefix}-{uuid}'.
    :rtype: IssueIdentifierResult
    :raises RuntimeError: If unable to generate unique ID after 10 attempts.
    """
    for _ in range(10):
        identifier = f"{request.prefix}-{_next_uuid_value()}"
        if identifier not in request.existing_ids:
            return IssueIdentifierResult(identifier=identifier)

    raise RuntimeError("unable to generate unique id after 10 attempts")


def generate_many_identifiers(title: str, prefix: str, count: int) -> Set[str]:
    """Generate multiple identifiers for uniqueness checks.

    :param title: Base title for hashing.
    :type title: str
    :param prefix: ID prefix.
    :type prefix: str
    :param count: Number of IDs to generate.
    :type count: int
    :return: Set of generated identifiers.
    :rtype: Set[str]
    """
    existing: Set[str] = set()
    for _ in range(count):
        request = IssueIdentifierRequest(
            title=title, prefix=prefix, existing_ids=existing
        )
        result = generate_issue_identifier(request)
        existing.add(result.identifier)
    return existing



"""Snyk vulnerability synchronization support.

Pulls vulnerabilities from the Snyk REST API and creates/updates Kanbus issues
under a configured parent epic.  The secret is read from the environment
variable SNYK_TOKEN (OAuth bearer token or legacy API key).

Hierarchy created:
  epic  "Snyk Vulnerabilities"
    task  "<repo>:<manifest-file>"  (one per affected file)
      sub-task  "[Snyk] KEY in pkg@version"  (one per vulnerability)
"""

from __future__ import annotations

import os
import subprocess
import logging
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set

import requests

from kanbus.ids import IssueIdentifierRequest, generate_issue_identifier
from kanbus.issue_files import (
    list_issue_identifiers,
    read_issue_from_file,
    write_issue_to_file,
)
from kanbus.models import IssueData, SnykConfiguration

SNYK_API_BASE = "https://api.snyk.io"
SNYK_API_VERSION = "2025-11-05"
SNYK_INITIATIVE_TITLE = "Snyk Vulnerability Remediation"
SNYK_DEP_EPIC_TITLE = "Snyk Dependency Vulnerabilities"
SNYK_CODE_EPIC_TITLE = "Snyk Code Vulnerabilities"


class SnykSyncError(RuntimeError):
    """Raised when a Snyk sync operation fails."""


@dataclass
class SnykPullResult:
    """Result of a Snyk pull operation."""

    pulled: int = 0
    updated: int = 0
    skipped: int = 0


def pull_from_snyk(
    root: Path,
    snyk_config: SnykConfiguration,
    project_key: str,
    dry_run: bool = False,
) -> SnykPullResult:
    """Pull vulnerabilities from Snyk and create/update Kanbus issues.

    Creates the hierarchy: epic → task-per-file → sub-task-per-vulnerability.

    :param root: Repository root path.
    :param snyk_config: Snyk configuration from .kanbus.yml.
    :param project_key: Kanbus project key (issue ID prefix).
    :param dry_run: If True, print what would be done without writing files.
    :raises SnykSyncError: If authentication or API calls fail.
    """
    token = os.environ.get("SNYK_TOKEN")
    if not token:
        raise SnykSyncError("SNYK_TOKEN environment variable is not set")

    from kanbus.project import load_project_directory

    project_dir = load_project_directory(root)
    issues_dir = project_dir / "issues"

    if not issues_dir.exists():
        raise SnykSyncError("issues directory does not exist")

    min_priority = _severity_to_priority(snyk_config.min_severity)

    # Resolve repo filter: use config value or auto-detect from git remote
    repo_filter = snyk_config.repo or _detect_repo_from_git(root)

    # Fetch projects map: project_id → target_file (filtered to this repo)
    project_map = _fetch_snyk_projects(snyk_config.org_id, token, repo_filter)

    # Fetch all vulnerabilities
    vulns = _fetch_all_snyk_issues(
        snyk_config.org_id, token, min_priority, ["package_vulnerability", "code"]
    )

    # Fetch enrichment data (fixedIn, description, cvssScore, etc.) from v1 API
    enrichment = _fetch_v1_enrichment(
        snyk_config.org_id, token, list(project_map.keys())
    )

    # Resolve or auto-create the parent epic(s) after we know which categories exist.
    all_existing: Set[str] = set(list_issue_identifiers(issues_dir))

    # Build indexes for idempotency
    snyk_key_index = _build_snyk_key_index(all_existing, issues_dir)
    file_task_index = _build_file_task_index(all_existing, issues_dir)

    # Group vulnerabilities by category + target_file, deduplicating by (project_id, key).
    file_to_vulns: Dict[tuple[str, str], List[Dict[str, Any]]] = defaultdict(list)
    has_code = False
    has_dependency = False
    category_priority: Dict[str, int] = {}
    seen_proj_key: set[str] = set()
    for vuln in vulns:
        proj_id = (
            vuln.get("relationships", {})
            .get("scan_item", {})
            .get("data", {})
            .get("id", "")
        )
        target_file = project_map.get(proj_id)
        if target_file is None:
            continue  # skip issues not in filtered project set
        issue_type = vuln.get("attributes", {}).get("type", "package_vulnerability")
        if issue_type == "code":
            category = "code"
            has_code = True
        else:
            category = "dependency"
            has_dependency = True
        priority = _severity_to_priority(
            vuln.get("attributes", {}).get("effective_severity_level", "low")
        )
        if category not in category_priority or priority < category_priority[category]:
            category_priority[category] = priority
        key = vuln.get("attributes", {}).get("key", "")
        dedup_key = f"{proj_id}:{key}"
        if dedup_key in seen_proj_key:
            continue
        seen_proj_key.add(dedup_key)
        file_to_vulns[(category, target_file)].append(vuln)

    epic_ids = _resolve_snyk_epics(
        issues_dir,
        project_key,
        snyk_config.parent_epic,
        dry_run,
        all_existing,
        include_dependency=has_dependency,
        include_code=has_code,
        dependency_priority=category_priority.get("dependency"),
        code_priority=category_priority.get("code"),
    )

    counts = SnykPullResult()

    for (category, target_file), file_vulns in sorted(file_to_vulns.items()):
        epic_id = epic_ids.get(category) or epic_ids.get("dependency")
        if not epic_id:
            continue
        # Highest priority (lowest number) among this file's vulnerabilities
        file_priority = min(
            (
                _severity_to_priority(
                    v.get("attributes", {}).get("effective_severity_level", "low")
                )
                for v in file_vulns
            ),
            default=3,
        )

        # Resolve or create a task for this file
        task_id = _resolve_file_task(
            issues_dir,
            project_key,
            target_file,
            category,
            FileTaskContext(epic_id=epic_id, priority=file_priority, dry_run=dry_run),
            file_task_index,
            all_existing,
        )

        # Create/update sub-tasks for each vulnerability in this file
        for vuln in file_vulns:
            snyk_key = _vuln_key(vuln)

            existing_kanbus_id = snyk_key_index.get(snyk_key)
            if existing_kanbus_id:
                kanbus_id = existing_kanbus_id
                action = "updated"
            else:
                request = IssueIdentifierRequest(
                    title=_vuln_title(vuln),
                    existing_ids=frozenset(all_existing),
                    prefix=project_key,
                )
                result = generate_issue_identifier(request)
                kanbus_id = result.identifier
                all_existing.add(kanbus_id)
                action = "pulled "

            v1_data = enrichment.get(snyk_key)
            issue = _map_snyk_to_kanbus(vuln, task_id, v1_data, target_file, root)
            issue = issue.model_copy(update={"identifier": kanbus_id})

            # Preserve created_at for updates
            issue_path = _issue_path(issues_dir, kanbus_id)
            if action == "updated" and issue_path.exists():
                try:
                    existing = read_issue_from_file(issue_path)
                    issue = issue.model_copy(update={"created_at": existing.created_at})
                except Exception as exc:
                    print(
                        f"Warning: failed to read existing issue file '{issue_path}' "
                        f"to preserve created_at: {exc}"
                    )

            short_key = (
                kanbus_id[: kanbus_id.find("-") + 7]
                if "-" in kanbus_id
                else kanbus_id[:6]
            )
            severity = vuln.get("attributes", {}).get("effective_severity_level", "?")
            print(f'{action}  [{severity:<8}]  {short_key:<14}  "{issue.title}"')

            if not dry_run:
                write_issue_to_file(issue, issue_path)

            if action == "updated":
                counts.updated += 1
            else:
                counts.pulled += 1

    return counts


def _resolve_parent_epic(
    issues_dir: Path,
    project_key: str,
    configured_id: Optional[str],
    dry_run: bool,
    all_existing: Set[str],
) -> str:
    """Resolve the parent epic ID, creating one if it doesn't exist."""
    if configured_id:
        path = _issue_path(issues_dir, configured_id)
        if path.exists():
            return configured_id

    existing_id = _find_existing_snyk_epic(issues_dir, all_existing)
    if existing_id:
        return existing_id

    request = IssueIdentifierRequest(
        title="Snyk Vulnerabilities",
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    result = generate_issue_identifier(request)
    epic_id = result.identifier
    all_existing.add(epic_id)

    now = datetime.now(timezone.utc)
    epic = IssueData.model_validate(
        {
            "id": epic_id,
            "title": "Snyk Vulnerabilities",
            "description": "Security vulnerabilities imported from Snyk.",
            "type": "epic",
            "status": "open",
            "priority": 1,
            "assignee": None,
            "creator": None,
            "parent": None,
            "labels": ["security", "snyk"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {},
        }
    )

    print(f'created  [epic    ]  {epic_id:<14}  "Snyk Vulnerabilities"')

    if not dry_run:
        write_issue_to_file(epic, _issue_path(issues_dir, epic_id))

    return epic_id


def _resolve_snyk_epics(
    issues_dir: Path,
    project_key: str,
    configured_id: Optional[str],
    dry_run: bool,
    all_existing: Set[str],
    include_dependency: bool,
    include_code: bool,
    dependency_priority: Optional[int],
    code_priority: Optional[int],
) -> Dict[str, str]:
    if configured_id:
        epic_id = _resolve_parent_epic(
            issues_dir, project_key, configured_id, dry_run, all_existing
        )
        epics: Dict[str, str] = {}
        if include_dependency:
            epics["dependency"] = epic_id
        if include_code:
            epics["code"] = epic_id
        return epics

    if not include_dependency and not include_code:
        return {}

    initiative_id = _resolve_snyk_initiative(
        issues_dir, project_key, dry_run, all_existing
    )
    epics: Dict[str, str] = {}
    if include_dependency:
        dep_epic = _resolve_snyk_epic(
            issues_dir,
            project_key,
            initiative_id,
            SNYK_DEP_EPIC_TITLE,
            "dependency",
            dependency_priority if dependency_priority is not None else 2,
            dry_run,
            all_existing,
        )
        epics["dependency"] = dep_epic
    if include_code:
        code_epic = _resolve_snyk_epic(
            issues_dir,
            project_key,
            initiative_id,
            SNYK_CODE_EPIC_TITLE,
            "code",
            code_priority if code_priority is not None else 2,
            dry_run,
            all_existing,
        )
        epics["code"] = code_epic
    return epics


def _resolve_snyk_initiative(
    issues_dir: Path,
    project_key: str,
    dry_run: bool,
    all_existing: Set[str],
) -> str:
    existing = _find_existing_snyk_initiative(issues_dir, all_existing)
    if existing:
        return existing

    request = IssueIdentifierRequest(
        title=SNYK_INITIATIVE_TITLE,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    result = generate_issue_identifier(request)
    initiative_id = result.identifier
    all_existing.add(initiative_id)

    now = datetime.now(timezone.utc)
    initiative = IssueData.model_validate(
        {
            "id": initiative_id,
            "title": SNYK_INITIATIVE_TITLE,
            "description": "Track remediation of Snyk vulnerabilities.",
            "type": "initiative",
            "status": "open",
            "priority": 1,
            "assignee": None,
            "creator": None,
            "parent": None,
            "labels": ["security", "snyk"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {},
        }
    )

    print(f'created  [initiative]  {initiative_id:<14}  "{SNYK_INITIATIVE_TITLE}"')

    if not dry_run:
        write_issue_to_file(initiative, _issue_path(issues_dir, initiative_id))

    return initiative_id


def _resolve_snyk_epic(
    issues_dir: Path,
    project_key: str,
    parent_initiative: str,
    title: str,
    category: str,
    priority: int,
    dry_run: bool,
    all_existing: Set[str],
) -> str:
    existing = _find_existing_snyk_epic(
        issues_dir, all_existing, title, parent_initiative
    )
    if existing:
        path = _issue_path(issues_dir, existing)
        issue = read_issue_from_file(path)
        if issue.priority != priority:
            issue.priority = priority
            issue.updated_at = datetime.now(timezone.utc)
            if not dry_run:
                write_issue_to_file(issue, path)
        return existing

    request = IssueIdentifierRequest(
        title=title,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    result = generate_issue_identifier(request)
    epic_id = result.identifier
    all_existing.add(epic_id)

    now = datetime.now(timezone.utc)
    epic = IssueData.model_validate(
        {
            "id": epic_id,
            "title": title,
            "description": "Security vulnerabilities imported from Snyk.",
            "type": "epic",
            "status": "open",
            "priority": priority,
            "assignee": None,
            "creator": None,
            "parent": parent_initiative,
            "labels": ["security", "snyk", category],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {"snyk_category": category},
        }
    )

    print(f'created  [epic    ]  {epic_id:<14}  "{title}"')

    if not dry_run:
        write_issue_to_file(epic, _issue_path(issues_dir, epic_id))

    return epic_id


def _find_existing_snyk_initiative(
    issues_dir: Path, all_existing: Set[str]
) -> Optional[str]:
    best_id: Optional[str] = None
    best_updated: Optional[datetime] = None

    for identifier in all_existing:
        issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        if issue.issue_type != "initiative":
            continue
        if issue.title != SNYK_INITIATIVE_TITLE:
            continue
        if "snyk" not in issue.labels:
            continue
        if best_updated is None or issue.updated_at > best_updated:
            best_updated = issue.updated_at
            best_id = issue.identifier

    return best_id


def _find_existing_snyk_epic(
    issues_dir: Path, all_existing: Set[str], title: str, parent_initiative: str
) -> Optional[str]:
    best_id: Optional[str] = None
    best_updated: Optional[datetime] = None

    for identifier in all_existing:
        issue = read_issue_from_file(_issue_path(issues_dir, identifier))
        if issue.issue_type != "epic":
            continue
        if issue.title != title:
            continue
        if "snyk" not in issue.labels:
            continue
        if issue.parent != parent_initiative:
            continue
        if best_updated is None or issue.updated_at > best_updated:
            best_updated = issue.updated_at
            best_id = issue.identifier

    return best_id


@dataclass
class FileTaskContext:
    epic_id: str
    priority: int
    dry_run: bool


def _resolve_file_task(
    issues_dir: Path,
    project_key: str,
    target_file: str,
    category: str,
    ctx: FileTaskContext,
    file_task_index: Dict[tuple[str, str], str],
    all_existing: Set[str],
) -> str:
    """Resolve or create a task for a manifest file under the epic."""
    key = (category, target_file)
    if key in file_task_index:
        task_id = file_task_index[key]
        if not ctx.dry_run:
            task_path = _issue_path(issues_dir, task_id)
            task = read_issue_from_file(task_path)
            if task.issue_type == "task":
                changed = False
                if task.parent != ctx.epic_id:
                    task.parent = ctx.epic_id
                    changed = True
                if task.priority != ctx.priority:
                    task.priority = ctx.priority
                    changed = True
                task.custom["snyk_target_file"] = target_file
                task.custom["snyk_category"] = category
                if "snyk" not in task.labels:
                    task.labels.append("snyk")
                    changed = True
                if "security" not in task.labels:
                    task.labels.append("security")
                    changed = True
                if changed:
                    task.updated_at = datetime.now(timezone.utc)
                    write_issue_to_file(task, task_path)
                    short_key = (
                        task_id[: task_id.find("-") + 7]
                        if "-" in task_id
                        else task_id[:6]
                    )
                    print(f'updated  [task    ]  {short_key:<14}  "{target_file}"')
        return task_id

    request = IssueIdentifierRequest(
        title=target_file,
        existing_ids=frozenset(all_existing),
        prefix=project_key,
    )
    result = generate_issue_identifier(request)
    task_id = result.identifier
    all_existing.add(task_id)

    now = datetime.now(timezone.utc)
    task = IssueData.model_validate(
        {
            "id": task_id,
            "title": target_file,
            "description": f"Snyk vulnerabilities found in `{target_file}`.",
            "type": "task",
            "status": "open",
            "priority": ctx.priority,
            "assignee": None,
            "creator": None,
            "parent": ctx.epic_id,
            "labels": ["security", "snyk"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": {"snyk_target_file": target_file, "snyk_category": category},
        }
    )

    short_key = task_id[: task_id.find("-") + 7] if "-" in task_id else task_id[:6]
    print(f'created  [task    ]  {short_key:<14}  "{target_file}"')

    if not ctx.dry_run:
        write_issue_to_file(task, _issue_path(issues_dir, task_id))

    return task_id


def _detect_repo_from_git(root: Path) -> Optional[str]:
    """Detect the GitHub repo slug from git remote origin URL.

    Returns e.g. "AnthusAI/Plexus" from https or SSH remote URLs.
    """
    try:
        result = subprocess.run(
            ["git", "remote", "get-url", "origin"],
            cwd=root,
            capture_output=True,
            text=True,
            timeout=5,
        )
        url = result.stdout.strip()
        if url.startswith("https://github.com/"):
            return url[len("https://github.com/") :].removesuffix(".git")
        if url.startswith("git@github.com:"):
            return url[len("git@github.com:") :].removesuffix(".git")
    except Exception as exc:
        # Best-effort detection; if git is unavailable or misconfigured, just return None.
        logging.getLogger(__name__).debug(
            "Failed to detect GitHub repo from git remote for %s: %s", root, exc
        )
    return None


def _fetch_snyk_projects(
    org_id: str, token: str, repo_filter: Optional[str] = None
) -> Dict[str, str]:
    """Fetch all projects, returning a map of project_id → target_file.

    If ``repo_filter`` is set (e.g. "AnthusAI/Plexus"), only projects whose
    Snyk name starts with ``"{repo_filter}:"`` are included.
    """
    headers = {
        "Authorization": f"Bearer {token}",
        "Accept": "application/vnd.api+json",
    }
    project_map: Dict[str, str] = {}
    prefix = f"{repo_filter}:" if repo_filter else None

    url: Optional[str] = (
        f"{SNYK_API_BASE}/rest/orgs/{org_id}/projects"
        f"?version={SNYK_API_VERSION}&limit=100"
    )

    while url:
        response = requests.get(url, headers=headers, timeout=30)
        if not response.ok:
            raise SnykSyncError(
                f"Snyk projects API returned {response.status_code}: {response.text[:200]}"
            )

        data = response.json()
        for project in data.get("data", []):
            proj_id = project.get("id", "")
            attrs = project.get("attributes", {})
            name = attrs.get("name", "")
            if prefix:
                repo = prefix.rstrip(":")
                if name != repo and not name.startswith(prefix):
                    continue
            target_file = attrs.get("target_file") or name
            if proj_id and target_file:
                project_map[proj_id] = target_file

        next_link = data.get("links", {}).get("next")
        url = f"{SNYK_API_BASE}{next_link}" if next_link else None

    return project_map


def _fetch_all_snyk_issues(
    org_id: str,
    token: str,
    min_priority: int,
    issue_types: Iterable[str],
) -> List[Dict[str, Any]]:
    """Fetch all issues from the Snyk REST API, filtering by min_priority.

    Returns all occurrences (one per project) without deduplication by key,
    so the same vulnerability may appear for multiple files.
    """
    all_issues: List[Dict[str, Any]] = []
    for issue_type in issue_types:
        try:
            all_issues.extend(
                _fetch_snyk_issues_for_type(org_id, token, min_priority, issue_type)
            )
        except SnykSyncError as exc:
            if issue_type == "package_vulnerability":
                raise
            logging.getLogger(__name__).warning(
                "Failed to fetch Snyk issues type '%s': %s", issue_type, exc
            )

    return all_issues


def _fetch_snyk_issues_for_type(
    org_id: str,
    token: str,
    min_priority: int,
    issue_type: str,
) -> List[Dict[str, Any]]:
    headers = {
        "Authorization": f"Bearer {token}",
        "Accept": "application/vnd.api+json",
    }
    all_issues: List[Dict[str, Any]] = []

    url: Optional[str] = (
        f"{SNYK_API_BASE}/rest/orgs/{org_id}/issues"
        f"?version={SNYK_API_VERSION}&limit=100&type={issue_type}"
    )

    while url:
        response = requests.get(url, headers=headers, timeout=30)
        if not response.ok:
            raise SnykSyncError(
                f"Snyk API returned {response.status_code}: {response.text[:200]}"
            )

        data = response.json()
        for issue in data.get("data", []):
            key = issue.get("attributes", {}).get("key", "")
            if not key:
                continue
            sev = issue.get("attributes", {}).get("effective_severity_level", "low")
            if _severity_to_priority(sev) <= min_priority:
                all_issues.append(issue)

        next_link = data.get("links", {}).get("next")
        url = f"{SNYK_API_BASE}{next_link}" if next_link else None

    return all_issues


def _fetch_v1_enrichment(
    org_id: str, token: str, project_ids: List[str]
) -> Dict[str, Any]:
    """Fetch enrichment data from the Snyk v1 aggregated-issues API.

    Returns a map of snyk_key → issue dict with fixedIn, description, cvssScore, etc.
    """
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
    enrichment: Dict[str, Any] = {}

    for proj_id in project_ids:
        url = f"{SNYK_API_BASE}/api/v1/org/{org_id}/project/{proj_id}/aggregated-issues"
        try:
            r = requests.post(
                url,
                headers=headers,
                json={"filters": {}, "includeDescription": True},
                timeout=30,
            )
            if not r.ok:
                continue
            for issue in r.json().get("issues", []):
                key = issue.get("issueData", {}).get("id", "")
                if key:
                    enrichment[key] = issue
        except Exception:
            continue

    return enrichment


def _build_snyk_key_index(existing_ids: Set[str], issues_dir: Path) -> Dict[str, str]:
    """Build a map from snyk_key → kanbus identifier."""
    index: Dict[str, str] = {}
    for identifier in existing_ids:
        path = _issue_path(issues_dir, identifier)
        try:
            issue = read_issue_from_file(path)
            snyk_key = issue.custom.get("snyk_key")
            if isinstance(snyk_key, str):
                index[snyk_key] = identifier
        except Exception as exc:
            logging.warning("Failed to read issue file %s: %s", path, exc)
    return index


def _build_file_task_index(
    existing_ids: Set[str], issues_dir: Path
) -> Dict[tuple[str, str], str]:
    """Build a map from (category, target_file) → kanbus task identifier."""
    index: Dict[tuple[str, str], str] = {}
    for identifier in existing_ids:
        path = _issue_path(issues_dir, identifier)
        try:
            issue = read_issue_from_file(path)
            if issue.issue_type == "task":
                target_file = issue.custom.get("snyk_target_file")
                if isinstance(target_file, str):
                    category = issue.custom.get("snyk_category")
                    if not isinstance(category, str):
                        category = "dependency"
                    index[(category, target_file)] = identifier
        except Exception as exc:
            logging.warning("Failed to read issue file %s: %s", path, exc)
    return index


def _vuln_key(issue: Dict[str, Any]) -> str:
    return issue.get("attributes", {}).get("key", "")


def _vuln_title(issue: Dict[str, Any]) -> str:
    attrs = issue.get("attributes", {})
    issue_type = attrs.get("type", "package_vulnerability")
    if issue_type == "code":
        title = attrs.get("title") or "Snyk Code issue"
        return f"[Snyk Code] {title}"
    key = attrs.get("key", "unknown")
    coords = attrs.get("coordinates", [{}])
    reps = coords[0].get("representations", [{}]) if coords else [{}]
    pkg = reps[0].get("dependency", {}).get("package_name", "") if reps else ""
    return f"[Snyk] {key} in {pkg}" if pkg else f"[Snyk] {key}"


def _extract_source_location(issue: Dict[str, Any]) -> Optional[Dict[str, object]]:
    coords = issue.get("attributes", {}).get("coordinates", [])
    for coord in coords:
        for rep in coord.get("representations", []):
            loc = rep.get("source_location") or rep.get("sourceLocation")
            if not loc:
                continue
            file_path = loc.get("file")
            if not file_path:
                continue
            region = loc.get("region")
            if region:
                start = region.get("start") or region
                end = region.get("end") or region
                return {
                    "file": file_path,
                    "line": start.get("line"),
                    "column": start.get("column"),
                    "end_line": end.get("line"),
                    "end_column": end.get("column"),
                }
            return {
                "file": file_path,
                "line": loc.get("line"),
                "column": loc.get("column") or loc.get("col"),
            }
    return None


def _extract_classes(issue: Dict[str, Any]) -> List[str]:
    classes = issue.get("attributes", {}).get("classes", [])
    parsed: List[str] = []
    for entry in classes:
        if isinstance(entry, str):
            parsed.append(entry)
            continue
        if isinstance(entry, dict):
            source = entry.get("source")
            ident = entry.get("id")
            if source and ident:
                parsed.append(f"{source}-{ident}")
    return parsed


SNIPPET_CONTEXT = 2
MAX_SNIPPET_LINES = 25


def _build_snippet(
    repo_root: Path,
    file_path: str,
    start_line: Optional[int],
    end_line: Optional[int],
) -> str:
    if not start_line or start_line <= 0:
        return ""
    if not end_line or end_line <= 0:
        end_line = start_line

    path = repo_root / file_path
    try:
        content = path.read_text(encoding="utf-8")
    except Exception:
        return ""

    lines = content.splitlines()
    if not lines:
        return ""

    total = len(lines)
    snippet_start = max(1, start_line - SNIPPET_CONTEXT)
    snippet_end = min(total, end_line + SNIPPET_CONTEXT)
    if snippet_end - snippet_start + 1 > MAX_SNIPPET_LINES:
        snippet_start = max(1, start_line - SNIPPET_CONTEXT)
        snippet_end = min(total, snippet_start + MAX_SNIPPET_LINES - 1)

    body_lines = []
    for line_no in range(snippet_start, snippet_end + 1):
        body_lines.append(f"{line_no:>4} | {lines[line_no - 1]}")

    body = "\n".join(body_lines)
    return (
        f"### Snippet ({file_path}:{snippet_start}-{snippet_end})\n```\n{body}\n```\n\n"
    )


def _map_snyk_to_kanbus(
    issue: Dict[str, Any],
    parent_task_id: Optional[str],
    v1: Optional[Dict[str, Any]] = None,
    target_file: str = "",
    repo_root: Optional[Path] = None,
) -> IssueData:
    """Map a Snyk issue dict to a Kanbus IssueData sub-task."""
    attrs = issue.get("attributes", {})
    issue_type = attrs.get("type", "package_vulnerability")
    snyk_key = attrs.get("key", "")
    severity = attrs.get("effective_severity_level", "low")
    priority = _severity_to_priority(severity)

    now = datetime.now(timezone.utc)
    custom: Dict[str, object] = {
        "snyk_key": snyk_key,
        "snyk_severity": severity,
        "snyk_type": issue_type,
    }

    if issue_type == "code":
        issue_title = attrs.get("title") or snyk_key
        description_text = attrs.get("description") or ""
        classes = _extract_classes(issue)
        class_line = f"**Classes:** {', '.join(classes)}\n" if classes else ""
        location = _extract_source_location(issue)
        file_line = loc_line = ""
        snippet_block = ""
        if location:
            file_line = f"**File:** `{location['file']}`\n"
            if location.get("line") is not None:
                if location.get("column") is not None:
                    if (
                        location.get("end_line") is not None
                        and location.get("end_column") is not None
                    ):
                        loc_line = (
                            f"**Location:** line {location['line']}, "
                            f"column {location['column']} to line "
                            f"{location['end_line']}, column {location['end_column']}\n"
                        )
                    else:
                        loc_line = (
                            f"**Location:** line {location['line']}, "
                            f"column {location['column']}\n"
                        )
                else:
                    loc_line = f"**Location:** line {location['line']}\n"
            custom["snyk_file"] = location["file"]
            if location.get("line") is not None:
                custom["snyk_line"] = int(location["line"])
            if location.get("column") is not None:
                custom["snyk_column"] = int(location["column"])
            if location.get("end_line") is not None:
                custom["snyk_end_line"] = int(location["end_line"])
            if location.get("end_column") is not None:
                custom["snyk_end_column"] = int(location["end_column"])
            if repo_root is not None:
                snippet_block = _build_snippet(
                    repo_root,
                    location["file"],
                    location.get("line"),
                    location.get("end_line"),
                )

        title = f"[Snyk Code] {issue_title}"
        details_block = (
            f"### Details\n{description_text}\n\n" if description_text else ""
        )
        description = (
            f"## {issue_title}\n\n"
            f"**Severity:** {severity}\n"
            f"**Project:** `{target_file}`\n"
            f"{file_line}"
            f"{loc_line}"
            f"{class_line}"
            f"{snippet_block}"
            f"**Issue Key:** {snyk_key}\n\n"
            f"{details_block}"
            f"**Fix:** Review and remediate in code."
        )
    else:
        coords = attrs.get("coordinates", [{}])
        reps = coords[0].get("representations", [{}]) if coords else [{}]
        dep = reps[0].get("dependency", {}) if reps else {}
        pkg_name = dep.get("package_name", "")
        pkg_version = dep.get("package_version", "")

        title = (
            f"[Snyk] {snyk_key} in {pkg_name}@{pkg_version}"
            if pkg_name
            else f"[Snyk] {snyk_key}"
        )

        # Prefer v1 human-readable title
        issue_data = (v1 or {}).get("issueData", {})
        vuln_title = issue_data.get("title") or snyk_key

        cves = [p["id"] for p in attrs.get("problems", []) if p.get("source") == "NVD"]
        cve_lines = (
            "\n".join(
                f"- [{cve}](https://nvd.nist.gov/vuln/detail/{cve})" for cve in cves
            )
            if cves
            else "No CVE assigned."
        )

        coord0 = coords[0] if coords else {}
        fixed_in = (v1 or {}).get("fixInfo", {}).get("fixedIn", [])
        if fixed_in:
            versions = ", ".join(fixed_in)
            if coord0.get("is_upgradeable"):
                fix_advice = (
                    f"**Fix:** Upgrade `{pkg_name}` to version {versions} or later."
                )
            else:
                fix_advice = (
                    f"**Fix:** Pin `{pkg_name}` to version {versions} or later."
                )
        elif coord0.get("is_upgradeable"):
            fix_advice = f"**Fix:** Upgrade `{pkg_name}` to a patched version."
        elif coord0.get("is_pinnable"):
            fix_advice = f"**Fix:** Pin `{pkg_name}` to a patched version."
        elif coord0.get("is_fixable_snyk"):
            fix_advice = "**Fix:** Snyk fix available — run `snyk fix`."
        else:
            fix_advice = "**Fix:** No automatic fix available. Review manually."

        # Extra metadata from v1
        cvss_score = issue_data.get("cvssScore")
        exploit_maturity = issue_data.get("exploitMaturity", "")
        priority_score = (v1 or {}).get("priorityScore")
        v1_description = issue_data.get("description", "")

        meta_lines = []
        if cvss_score is not None:
            meta_lines.append(f"**CVSS Score:** {cvss_score:.1f}")
        if exploit_maturity and exploit_maturity != "no-known-exploit":
            meta_lines.append(f"**Exploit Maturity:** {exploit_maturity}")
        if priority_score is not None:
            meta_lines.append(f"**Snyk Priority Score:** {priority_score}/1000")
        meta_block = ("  \n".join(meta_lines) + "\n\n") if meta_lines else ""

        detail_block = f"### Details\n{v1_description}\n\n" if v1_description else ""

        snyk_url = f"https://security.snyk.io/vuln/{snyk_key}"
        description = (
            f"## {vuln_title}\n\n"
            f"**Severity:** {severity}\n"
            f"**Package:** {pkg_name}@{pkg_version}\n"
            f"**File:** `{target_file}`\n\n"
            f"{meta_block}"
            f"### CVEs\n{cve_lines}\n\n"
            f"{fix_advice}\n\n"
            f"{detail_block}"
            f"### Reference\n- [Snyk advisory]({snyk_url})"
        )

    return IssueData.model_validate(
        {
            "id": "__placeholder__",
            "title": title,
            "description": description,
            "type": "sub-task",
            "status": "open",
            "priority": priority,
            "assignee": None,
            "creator": None,
            "parent": parent_task_id,
            "labels": ["security", "snyk"],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": custom,
        }
    )


def _severity_to_priority(severity: str) -> int:
    """Map Snyk severity to Kanbus priority. critical=0, high=1, medium=2, low=3."""
    return {"critical": 0, "high": 1, "medium": 2}.get(severity.lower(), 3)


def _issue_path(issues_dir: Path, identifier: str) -> Path:
    return issues_dir / f"{identifier}.json"



"""Issue close workflow."""

from __future__ import annotations

from pathlib import Path

from kanbus.issue_update import IssueUpdateError, update_issue
from kanbus.models import IssueData


class IssueCloseError(RuntimeError):
    """Raised when issue closing fails."""


def close_issue(root: Path, identifier: str) -> IssueData:
    """Close an issue by transitioning it to closed status.

    :param root: Repository root path.
    :type root: Path
    :param identifier: Issue identifier.
    :type identifier: str
    :return: Updated issue data.
    :rtype: IssueData
    :raises IssueCloseError: If closing fails.
    """
    try:
        return update_issue(
            root=root,
            identifier=identifier,
            title=None,
            description=None,
            status="closed",
            assignee=None,
            claim=False,
            validate=True,
        )
    except IssueUpdateError as error:
        raise IssueCloseError(str(error)) from error



"""Event history recording and retrieval helpers."""

from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional
from uuid import uuid4

from pydantic import BaseModel, Field

from kanbus.models import IssueData

EVENT_SCHEMA_VERSION = 1


class EventRecord(BaseModel):
    """Event history record for a single issue action."""

    schema_version: int = Field(default=EVENT_SCHEMA_VERSION)
    event_id: str
    issue_id: str
    event_type: str
    occurred_at: str
    actor_id: str
    payload: Dict[str, Any]


def now_timestamp() -> str:
    """Return the current UTC timestamp formatted for filenames."""
    return (
        datetime.now(timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )


def event_filename(occurred_at: str, event_id: str) -> str:
    return f"{occurred_at}__{event_id}.json"


def create_event(
    issue_id: str,
    event_type: str,
    actor_id: str,
    payload: Dict[str, Any],
    occurred_at: Optional[str] = None,
) -> EventRecord:
    timestamp = occurred_at or now_timestamp()
    return EventRecord(
        event_id=str(uuid4()),
        issue_id=issue_id,
        event_type=event_type,
        occurred_at=timestamp,
        actor_id=actor_id,
        payload=payload,
    )


def events_dir_for_project(project_dir: Path) -> Path:
    return project_dir / "events"


def events_dir_for_local(project_dir: Path) -> Path:
    parent = project_dir.parent
    if parent is None:
        raise RuntimeError("project-local path unavailable")
    return parent / "project-local" / "events"


def events_dir_for_issue_path(project_dir: Path, issue_path: Path) -> Path:
    local_dir = project_dir.parent / "project-local"
    if local_dir.is_dir() and issue_path.is_relative_to(local_dir):
        return local_dir / "events"
    return events_dir_for_project(project_dir)


def events_dir_for_issue(project_dir: Path, issue_id: str) -> Path:
    local_dir = project_dir.parent / "project-local"
    if local_dir.is_dir():
        local_issue = local_dir / "issues" / f"{issue_id}.json"
        if local_issue.exists():
            return local_dir / "events"
    return events_dir_for_project(project_dir)


def write_events_batch(events_dir: Path, events: Iterable[EventRecord]) -> List[Path]:
    events_list = list(events)
    if not events_list:
        return []
    events_dir.mkdir(parents=True, exist_ok=True)
    written: List[Path] = []
    for event in events_list:
        filename = event_filename(event.occurred_at, event.event_id)
        final_path = events_dir / filename
        temp_path = events_dir / f".{filename}.tmp"
        try:
            with temp_path.open("x", encoding="utf-8") as handle:
                handle.write(event.model_dump_json(indent=2))
                handle.flush()
            temp_path.replace(final_path)
            written.append(final_path)
        except Exception as error:  # noqa: BLE001
            if temp_path.exists():
                temp_path.unlink(missing_ok=True)
            rollback_event_files(written)
            raise RuntimeError(str(error)) from error
    return written


def rollback_event_files(paths: Iterable[Path]) -> None:
    for path in paths:
        path.unlink(missing_ok=True)


def issue_created_payload(issue: IssueData) -> Dict[str, Any]:
    return {
        "title": issue.title,
        "description": issue.description,
        "issue_type": issue.issue_type,
        "status": issue.status,
        "priority": issue.priority,
        "assignee": issue.assignee,
        "parent": issue.parent,
        "labels": issue.labels,
    }


def issue_deleted_payload(issue: IssueData) -> Dict[str, Any]:
    return {
        "title": issue.title,
        "issue_type": issue.issue_type,
        "status": issue.status,
    }


def state_transition_payload(from_status: str, to_status: str) -> Dict[str, Any]:
    return {"from_status": from_status, "to_status": to_status}


def comment_payload(comment_id: str, comment_author: str) -> Dict[str, Any]:
    return {"comment_id": comment_id, "comment_author": comment_author}


def comment_updated_payload(comment_id: str, comment_author: str) -> Dict[str, Any]:
    return {
        "comment_id": comment_id,
        "comment_author": comment_author,
        "changed_fields": ["text"],
    }


def dependency_payload(dependency_type: str, target_id: str) -> Dict[str, Any]:
    return {"dependency_type": dependency_type, "target_id": target_id}


def transfer_payload(from_location: str, to_location: str) -> Dict[str, Any]:
    return {"from_location": from_location, "to_location": to_location}


def field_update_payload(
    before: IssueData, after: IssueData
) -> Optional[Dict[str, Any]]:
    changes: Dict[str, Dict[str, Any]] = {}

    def push(field: str, from_value: Any, to_value: Any) -> None:
        if from_value == to_value:
            return
        changes[field] = {"from": from_value, "to": to_value}

    push("title", before.title, after.title)
    push("description", before.description, after.description)
    push("assignee", before.assignee, after.assignee)
    push("priority", before.priority, after.priority)
    push("labels", before.labels, after.labels)
    push("parent", before.parent, after.parent)

    if not changes:
        return None
    return {"changes": changes}


def build_update_events(
    before: IssueData, after: IssueData, actor_id: str, occurred_at: str
) -> List[EventRecord]:
    events: List[EventRecord] = []
    if before.status != after.status:
        events.append(
            create_event(
                issue_id=after.identifier,
                event_type="state_transition",
                actor_id=actor_id,
                payload=state_transition_payload(before.status, after.status),
                occurred_at=occurred_at,
            )
        )
    payload = field_update_payload(before, after)
    if payload is not None:
        events.append(
            create_event(
                issue_id=after.identifier,
                event_type="field_updated",
                actor_id=actor_id,
                payload=payload,
                occurred_at=occurred_at,
            )
        )
    return events
